import sys

import click as click

from oarepo_cli.cli.utils import with_config
from oarepo_cli.utils import run_cmdline


@click.command(name="run", help="Run the server")
@click.option("-c", "--celery")
@click.argument("site", default=None, required=False)
@with_config()
def run_server(cfg=None, celery=False, site=None, **kwargs):
    sites = cfg.whole_data.get("sites", {})
    if not site:
        if len(sites) == 1:
            site = next(iter(sites.keys()))
        else:
            print(
                f"You have more than one site installed ({list(sites.keys())}), please specify its name on the commandline"
            )
            sys.exit(1)
    else:
        if site not in sites:
            print(
                f"Site with name {site} not found in repository sites {list(sites.keys())}"
            )
            sys.exit(1)
    if celery:
        run_invenio_cli(cfg, sites[site])
    else:
        run_pipenv_server(cfg, sites[site])


def run_invenio_cli(config, site):
    invenio_cli = config.project_dir / config.get("invenio_cli")
    site_dir = config.project_dir.absolute() / site["site_dir"]
    run_cmdline(
        "pipenv",
        "run",
        invenio_cli,
        "run",
        cwd=site_dir,
        environ={"PIPENV_IGNORE_VIRTUALENVS": "1"},
    )


def run_pipenv_server(config, site):
    site_dir = config.project_dir.absolute() / site["site_dir"]
    run_cmdline(
        "pipenv",
        "run",
        "invenio",
        "run",
        "--cert",
        "docker/nginx/test.crt",
        "--key",
        "docker/nginx/test.key",
        cwd=site_dir,
        environ={"PIPENV_IGNORE_VIRTUALENVS": "1"},
    )



import logging

from git import Repo

from oarepo_cli.ui.wizard import WizardStep
from oarepo_cli.utils import to_python_name

log = logging.getLogger("step_01_initialize_directory")


class DirectoryStep(WizardStep):
    def __init__(self, *args, **kwargs):
        super().__init__(heading="Creating the target directory ...")

    def after_run(self):
        self.data["project_package"] = to_python_name(self.data.project_dir.name)
        p = self.data.project_dir
        if not p.exists():
            p.mkdir(parents=True)
        if not (p / ".git").exists():
            Repo.init(p)

    def should_run(self):
        return True



from oarepo_cli.cli.site.utils import SiteWizardStepMixin
from oarepo_cli.ui.wizard import WizardStep


class NextStepsStep(SiteWizardStepMixin, WizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            heading=lambda data: f"""
The repository skeleton has been created and dependencies installed.
To check that everything has been installed successfully you may want
to start your new repository by running

    nrp-cli run
    
and point your browser to 

    https://localhost:5000/
    
After doing so, add your model by calling

    nrp-cli model add <modelname>
    
edit the contents of metadata.yaml to suit your needs and run

    nrp-cli model compile <modelname>
    nrp-cli model install <modelname>
    
If you run invenio again and head to https://localhost:5000/api/<modelname>
an empty listing should be returned.

In the next step, add a UI application by calling

    nrp-cli ui add <uiname>
    nrp-cli ui compile <uiname>
    nrp-cli ui install <uiname>

After restarting the server, a UI will be available at 
https://localhost:5000/<uiname>
            """,
            **kwargs,
        )

    def should_run(self):
        return True



from pathlib import Path
import shutil

from pkg_resources import parse_requirements

from oarepo_cli.cli.site.utils import SiteWizardStepMixin
from oarepo_cli.ui.radio import Radio
from oarepo_cli.ui.wizard import WizardStep

from ...utils import commit_git, run_cmdline


class CreatePipenvStep(SiteWizardStepMixin, WizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            Radio(
                "create_pipenv_in_site",
                options={
                    "yes": ".venv inside my site directory",
                    "no": "~/.config/virtualenvs/<mangled_name> (standard pipenv location)",
                },
                default="yes",
            ),
            heading="""
In this step I will create python environment for this repository.
Note that this can take a couple of minutes to finish
during which "Locking ..." will be displayed.

What is your preference of pipenv virtual environment location?
            """,
            **kwargs,
        )

    def after_run(self):
        site_dir = self.site_dir
        if self.data.get("create_pipenv_in_site") == "yes":
            (site_dir / ".venv").mkdir(parents=True, exist_ok=True)
        python_binary = shutil.which(self.data.whole_data["config"]["python"])
        run_cmdline(
            "pipenv",
            "lock",
            "--python",
            python_binary,
            cwd=site_dir,
            environ={"PIPENV_IGNORE_VIRTUALENVS": "1"},
        )
        run_cmdline(
            "pipenv",
            "install",
            "--python",
            python_binary,
            cwd=site_dir,
            environ={"PIPENV_IGNORE_VIRTUALENVS": "1"},
        )
        run_cmdline(
            "pipenv",
            "run",
            "which",
            "python",
            cwd=site_dir,
            environ={"PIPENV_IGNORE_VIRTUALENVS": "1"},
        )

        pipenv_venv_dir = Path(self._pipenv_venv_dir).relative_to(self.data.project_dir)

        self.data["site_pipenv_dir"] = str(pipenv_venv_dir)
        commit_git(
            self.data.project_dir,
            f"after-site-pipenv-{self.data.section}",
            f"Committed automatically after site {self.data.section} pipenv has been called",
        )

    @property
    def _pipenv_venv_dir(self):
        success = run_cmdline(
            "pipenv",
            "--venv",
            cwd=self.site_dir,
            environ={"PIPENV_IGNORE_VIRTUALENVS": "1"},
            check_only=True,
            grab_stdout=True,
        )
        if not success:
            return None
        return success.strip()

    def should_run(self):
        should_run = (
            "site_pipenv_dir" not in self.data
            or not (self.data.project_dir / self.data["site_pipenv_dir"]).exists()
        )
        if should_run:
            return should_run

        # the directory exists, fetch requirements and compare them
        pipfile_requirements = self.load_reqs(
            run_cmdline(
                "pipenv",
                "requirements",
                cwd=self.site_dir,
                environ={"PIPENV_IGNORE_VIRTUALENVS": "1"},
                check_only=True,
                grab_stdout=True,
            )
        )

        virtualenv_requirements = self.load_reqs(
            run_cmdline(
                "pipenv",
                "run",
                "pip",
                "freeze",
                cwd=self.site_dir,
                environ={"PIPENV_IGNORE_VIRTUALENVS": "1"},
                check_only=True,
                grab_stdout=True,
            )
        )

        IGNORED_PACKAGES = {"setuptools", "pip", "wheel"}

        requirement_not_found = False

        for key, pipfile_version in pipfile_requirements.items():
            virtualenv_version = virtualenv_requirements.get(key)
            if not virtualenv_version:
                if key in IGNORED_PACKAGES:
                    continue
                print(f"Requirement {key} not found, will run pipfile lock")
                requirement_not_found = True
            elif pipfile_version != virtualenv_version:
                print(
                    f"Requirement version mismatch in {key}. Expected {pipfile_version}, is {virtualenv_version}. Will run pipfile lock"
                )
                requirement_not_found = True

        return requirement_not_found

    def load_reqs(self, txt):
        ret = {}
        for l in txt.splitlines():
            l = l.strip()
            if l[0] == "-" or l[0] == "#":
                continue
            try:
                for resource in parse_requirements(l):
                    pn = resource.project_name.replace("_", "-").lower()
                    version = None
                    for spec in resource.specs:
                        if spec[0] == "==":
                            version = spec[1]
                    ret[pn] = version
            except Exception as e:
                print(f"Error parsing {l}: {e}")
        return ret



import venv

import click as click
from colorama import Fore, Style
import yaml

from oarepo_cli.cli.model.utils import ModelWizardStep
from oarepo_cli.cli.utils import with_config
from oarepo_cli.ui.wizard import Wizard, WizardStep
from oarepo_cli.ui.wizard.steps import RadioWizardStep
from oarepo_cli.utils import commit_git, pip_install, run_cmdline


@click.command(
    name="compile",
    help="""
Compile model yaml file to invenio sources. Required arguments:
    <name>   ... name of the already existing model
""",
)
@click.argument("name", required=True)
@with_config(config_section=lambda name, **kwargs: ["models", name])
def compile_model(cfg=None, **kwargs):
    commit_git(
        cfg.project_dir,
        f"before-model-compile-{cfg.section}",
        f"Committed automatically before model {cfg.section} has been compiled",
    )
    optional_steps = []
    model_dir = cfg.project_dir / "models" / cfg.section
    if (model_dir / "setup.cfg").exists():
        optional_steps.append(
            RadioWizardStep(
                "merge_changes",
                options={
                    "merge": "Merge changes into the previously generated files",
                    "overwrite": "Remove previously generated files and start from scratch",
                },
                default="merge",
                heading=f"""
It seems that the model has been already generated. 

Should I try to {Fore.GREEN}merge{Fore.BLUE} the changes with the existing sources 
or {Fore.RED}remove{Fore.BLUE} the previously generated sources and generate from scratch?

{Fore.YELLOW}Please make sure that you have your existing sources safely committed into git repository
so that you might recover them if the compilation process fails.{Style.RESET_ALL}
""",
            )
        )
    wizard = Wizard(*optional_steps, CompileWizardStep())
    wizard.run(cfg)
    commit_git(
        cfg.project_dir,
        f"after-model-compile-{cfg.section}",
        f"Committed automatically after model {cfg.section} has been compiled",
    )


class CompileWizardStep(ModelWizardStep, WizardStep):
    def after_run(self):
        venv_dir = self.data.project_dir / ".venv" / "oarepo-model-builder"
        venv_dir = venv_dir.absolute()
        if not venv_dir.exists():
            venv.main([str(venv_dir)])

        # TODO: install plugins - but note, there might be error parsing the file as some includes might be handled by the plugin
        pip_install(
            venv_dir / "bin" / "pip",
            "OAREPO_MODEL_BUILDER_VERSION",
            "oarepo-model-builder==3.*",
            "https://github.com/oarepo/oarepo-model-builder",
        )
        pip_install(
            venv_dir / "bin" / "pip",
            "OAREPO_MODEL_BUILDER_TESTS_VERSION",
            "oarepo-model-builder-tests==3.*",
            "https://github.com/oarepo/oarepo-model-builder-tests",
        )
        if self.data.get("use_requests", None) == "yes":
            pip_install(
                venv_dir / "bin" / "pip",
                "OAREPO_MODEL_BUILDER_REQUESTS_VERSION",
                "oarepo-model-builder-requests==3.*",
                "https://github.com/oarepo/oarepo-model-builder-requests",
            )
        if self.data.get("use_expandable_fields", None) == "yes":
            pip_install(
                venv_dir / "bin" / "pip",
                "OAREPO_MODEL_BUILDER_EXPANSIONS_VERSION",
                "oarepo-model-builder-expansions==3.*",
                "https://github.com/oarepo/oarepo-model-builder-expansions",
            )
        if self.data.get("use_files", None) == "yes":
            pip_install(
                venv_dir / "bin" / "pip",
                "OAREPO_MODEL_BUILDER_FILES_VERSION",
                "oarepo-model-builder-files==3.*",
                "https://github.com/oarepo/oarepo-model-builder-files",
            )

        if self.data.get("use_drafts", None) == "yes":
            pip_install(
                venv_dir / "bin" / "pip",
                "OAREPO_MODEL_BUILDER_DRAFTS_VERSION",
                "oarepo-model-builder-drafts==1.*",
                "https://github.com/oarepo/oarepo-model-builder-drafts",
            )

        if not self.data.get("use_drafts", None) == "yes":
            run_cmdline(
                venv_dir / "bin" / "pip",
                "uninstall",
                "-y",
                "oarepo-model-builder-drafts",
                cwd=self.model_dir,
            )

        if (
            self.data.get("use_relations", None) == "yes"
            or self.data.get("use_vocabularies", None) == "yes"
        ):
            pip_install(
                venv_dir / "bin" / "pip",
                "OAREPO_MODEL_BUILDER_RELATIONS_VERSION",
                "oarepo-model-builder-relations==1.*",
                "https://github.com/oarepo/oarepo-model-builder-relations",
            )
        if self.data.get("use_vocabularies", None) == "yes":
            pip_install(
                venv_dir / "bin" / "pip",
                "OAREPO_MODEL_BUILDER_VOCABULARIES_VERSION",
                "oarepo-model-builder-vocabularies==1.*",
                "https://github.com/oarepo/oarepo-model-builder-vocabularies",
            )
        if self.data.get("use_custom_fields", None) == "yes":
            pip_install(
                venv_dir / "bin" / "pip",
                "OAREPO_MODEL_BUILDER_FILES_VERSION",
                "oarepo-model-builder-cf==1.*",
                "https://github.com/oarepo/oarepo-model-builder-cf",
            )

        # TODO: install plugins - but note, there might be error parsing the file as some includes might be handled by the plugin
        with open(self.model_dir / "model.yaml") as f:
            model_data = yaml.safe_load(f)
        plugins = model_data.get("model", {}).get("plugins", {}).get("packages", [])
        for package in plugins:
            run_cmdline(
                venv_dir / "bin" / "pip",
                "install",
                package,
                cwd=self.model_dir,
            )

        opts = []

        if self.data.get("use_files", None) == "yes":
            opts.append("--profile")
            opts.append("model,files")

        if self.data.get("use_drafts", None) == "yes":
            opts.append("--profile")
            opts.append("model,drafts")

        if self.data.get("merge_changes", None) == "overwrite":
            opts.append("--overwrite")

        run_cmdline(
            venv_dir / "bin" / "oarepo-compile-model",
            *opts,
            "-vvv",
            "model.yaml",
            cwd=self.model_dir,
        )

    def should_run(self):
        # always run as there is an optional step for merge/overwrite changes
        return True



import subprocess

from oarepo_cli.cli.site.utils import SiteWizardStepMixin
from oarepo_cli.ui.wizard import WizardStep

from ...utils import run_cmdline


class StartContainersStep(SiteWizardStepMixin, WizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            heading="""
I'm going to start docker containers (database, opensearch, message queue, cache etc.).
If this step fails, please fix the problem and run the wizard again.            
            """,
            **kwargs
        )

    def after_run(self):
        run_cmdline(
            self.data.project_dir / self.data.get("config.invenio_cli"),
            "services",
            "start",
            cwd=self.site_dir,
            environ={"PIPENV_IGNORE_VIRTUALENVS": "1"},
        )
        self._check_containers_running(False)

    def _check_containers_running(self, check_only):
        try:
            stdout = run_cmdline(
                self.data.project_dir / self.data.get("config.invenio_cli"),
                "services",
                "status",
                cwd=self.site_dir,
                environ={"PIPENV_IGNORE_VIRTUALENVS": "1"},
                check_only=check_only,
                grab_stdout=True,
            )
            if not isinstance(stdout, str):
                return False
            if "unable to connect" in stdout:
                return False
        except subprocess.CalledProcessError:
            return False
        return True

    def should_run(self):
        return not self._check_containers_running(True)



from __future__ import annotations

import shutil
import venv
from pathlib import Path

from oarepo_cli.ui.wizard import WizardStep

from ...utils import commit_git, pip_install


class InstallInvenioCliStep(WizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            heading="""I will install invenio command-line client. I will use the client to check that
you have all the tools installed (python, docker, ...). If this step fails, please fix the problem
and run this wizard again.

Before the installation, you may have a look at the requirements at
https://inveniordm.docs.cern.ch/install/requirements/ .
            """,
            **kwargs,
        )

    def after_run(self):
        print("Creating invenio-cli virtualenv")
        invenio_cli_dir = self._invenio_cli_dir
        self.data["invenio_cli"] = str(
            (invenio_cli_dir / "bin" / "invenio-cli").relative_to(self.data.project_dir)
        )
        if invenio_cli_dir.exists():
            shutil.rmtree(invenio_cli_dir)
        venv.main([str(invenio_cli_dir)])

        pip_install(
            invenio_cli_dir / "bin" / "pip",
            "INVENIO_CLI_VERSION",
            "invenio-cli==1.0.16",
            "https://github.com/inveniosoftware/invenio-cli",
        )

        with open(invenio_cli_dir / ".install.ok", "w") as f:
            f.write("invenio installation ok")
        commit_git(
            self.data.project_dir,
            "after-install-invenio-cli",
            "Committed automatically after invenio-cli has been installed",
        )

    @property
    def _invenio_cli_dir(self):
        return Path(self.data.project_dir) / ".venv" / "invenio-cli"

    def should_run(self):
        return not (self._invenio_cli_dir / ".install.ok").exists()



import click

from oarepo_cli.cli.initialize import initialize
from oarepo_cli.cli.model import model
from oarepo_cli.cli.run import run_server
from oarepo_cli.cli.site import site
from oarepo_cli.cli.ui import ui
from oarepo_cli.cli.upgrade import upgrade
from oarepo_cli.cli.develop_docker import docker_develop
from oarepo_cli.cli.develop import develop


@click.group()
def run(*args, **kwargs):
    pass


run.add_command(initialize)
run.add_command(site)
run.add_command(model)
run.add_command(ui)
run.add_command(run_server)
run.add_command(upgrade)
run.add_command(docker_develop)
run.add_command(develop)

if __name__ == "__main__":
    run()



import click

from oarepo_cli.cli.site.add import add_site
from oarepo_cli.cli.utils import with_config
from oarepo_cli.utils import print_banner

from ...ui.wizard import StaticWizardStep, Wizard
from .step_01_initialize_directory import DirectoryStep
from .step_02_deployment_type import DeploymentTypeStep
from .step_03_create_monorepo import CreateMonorepoStep
from .step_04_install_invenio_cli import InstallInvenioCliStep
from .step_05_install_oarepo_cli import InstallIOARepoCliStep
from .step_06_primary_site_name import PrimarySiteNameStep


@click.command(
    name="initialize",
    help="""
Initialize the whole repository structure. Required arguments:
    <project_dir>   ... path to the output directory
""",
)
@click.option(
    "--no-site",
    default=False,
    is_flag=True,
    type=bool,
    help="Do not create default site",
)
@click.option("--python", required=False)
@with_config(project_dir_as_argument=True)
@click.pass_context
def initialize(ctx, cfg=None, no_site=None, python=None, **kwargs):
    if not kwargs["no_banner"]:
        print_banner()
    cfg["python"] = python or "python3.9"
    initialize_wizard = Wizard(
        StaticWizardStep(
            """
        This command will initialize a new repository based on OARepo codebase (an extension of Invenio repository).
                """,
            pause=True,
        ),
        DirectoryStep(),
        DeploymentTypeStep(),
        CreateMonorepoStep(pause=True),
        InstallInvenioCliStep(pause=True),
        InstallIOARepoCliStep(pause=True),
        PrimarySiteNameStep(),
    )
    initialize_wizard.run(cfg)

    # install all sites from the config file
    sites = cfg.whole_data.get("sites", {})
    if sites:
        for site in sites:
            ctx.invoke(
                add_site,
                project_dir=str(cfg.project_dir),
                name=site,
                no_banner=True,
                no_input=cfg.no_input,
                silent=cfg.silent,
            )
    elif not no_site:
        # if there is no site, install the default one
        ctx.invoke(
            add_site,
            project_dir=str(cfg.project_dir),
            name=cfg["primary_site_name"],
            no_banner=True,
            no_input=cfg.no_input,
            silent=cfg.silent,
        )
    # do not install model nor uis as user might (and will) modify them


if __name__ == "__main__":
    initialize()



from __future__ import annotations

from typing import Callable, Dict, Union

from .steps import WizardBase, WizardStep


class Wizard(WizardBase):
    def __init__(self, *steps: Union[WizardStep, Callable[[Dict], None], str]):
        super().__init__(steps)

    def should_run(self, data):
        return super().should_run(data)

    def run(self, data):
        super().run(data)
        self.after_run()

    def after_run(self):
        pass



import os

from oarepo_cli.cli.utils import ProjectWizardMixin, SiteMixin
from oarepo_cli.ui.wizard import WizardStep


class ModelWizardStep(SiteMixin, ProjectWizardMixin, WizardStep):
    @property
    def model_name(self):
        return self.data.section

    @property
    def model_dir(self):
        return self.data.project_dir / "models" / self.model_name

    @property
    def model_package_dir(self):
        return self.model_dir / os.sep.join(self.data["model_package"].split("."))



try:
    pass
except ImportError:
    pass

from typing import Dict

from colorama import Fore, Style

from oarepo_cli.ui.widget import Widget


class Radio(Widget):
    def __init__(self, name, options: Dict[str, str], default=None):
        super().__init__(name)
        self.options = options
        self.default = default

    def run(self):
        displayed = [
            (str(idx + 1), key, label)
            for idx, (key, label) in enumerate(self.options.items())
        ]
        print()
        for d in displayed:
            print(f"{Fore.YELLOW}{d[0]}{Style.RESET_ALL}) {d[2]}")

        value = self.value or self.default
        option = None
        for d in displayed:
            if d[1] == value:
                option = d[0]
                break
        if option:
            prompt = f"Your choice [{option}]: "
        else:
            prompt = "Your choice: "

        while True:
            print()
            value = input(prompt).strip() or option
            for d in displayed:
                if d[0] == value:
                    return d[1]
            print(
                f"{Fore.RED}Bad option: select one of the options above{Style.RESET_ALL}"
            )



import click as click

from .add import add_model
from .compile import compile_model
from .install import install_model
from .list import list_models
from .load import load_data


@click.group(
    help="Model-related tools (add model, compile, install, load and dump data)"
)
def model():
    pass


model.add_command(add_model)
model.add_command(compile_model)
model.add_command(install_model)
model.add_command(load_data)
model.add_command(list_models)



from __future__ import annotations

import os
from pathlib import Path

from oarepo_cli.cli.site.utils import SiteWizardStepMixin
from oarepo_cli.ui.wizard import WizardStep

from ...utils import run_cmdline


class CheckRequirementsStep(SiteWizardStepMixin, WizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            heading="""
I will check the requirements for Invenio site installation.
            """,
            **kwargs,
        )

    def after_run(self):
        invenio_cli_dir = self._invenio_cli_dir

        run_cmdline(
            invenio_cli_dir / "bin" / "invenio-cli",
            "check-requirements",
            "--development",
            environ={
                "PIPENV_IGNORE_VIRTUALENVS": "1",
                "PATH": f"{self.data.project_dir}/.bin:{os.environ['PATH']}",
            },
            cwd=self.site_dir,
        )
        with open(invenio_cli_dir / ".check.ok", "w") as f:
            f.write("invenio check ok")

    def should_run(self):
        return not (self._invenio_cli_dir / ".check.ok").exists()

    @property
    def _invenio_cli_dir(self):
        return Path(self.data.project_dir) / ".venv" / "invenio-cli"



from colorama import Fore, Style

try:
    import gnureadline as readline
except ImportError:
    import readline

from oarepo_cli.ui.widget import Widget


class Input(Widget):
    def __init__(self, name, default=None, prompt=None):
        super().__init__(name, default=default)
        self.prompt = prompt

    def run(self):
        def hook():
            readline.insert_text(self.value or "")
            readline.redisplay()

        readline.set_pre_input_hook(hook)

        try:
            prompt = self.prompt or "Enter value"
            line = input(f"{Fore.BLUE}{prompt}: {Style.RESET_ALL}")
        finally:
            readline.set_pre_input_hook()
        self.value = line
        return line



from __future__ import annotations

import abc
import copy
from typing import Any, Callable, Dict, List, Union

from colorama import Fore, Style

from ...config import Config
from ..input import Input
from ..radio import Radio
from ..utils import slow_print
from .validation import required as required_validation


class WizardBase(abc.ABC):
    steps: "List[Union[WizardStep, Callable[[Dict], None], str]]" = []
    data: Dict[str, Any] = None

    def __init__(
        self, steps: "List[Union[WizardStep, Callable[[Dict], None], str]]" = None
    ):
        self.steps = steps or self.steps

    def run(self, data):
        self.data = data
        steps = self.get_steps()
        for step in steps:
            step.data = data
        for stepidx, step in enumerate(steps):
            should_run = step.should_run()
            if should_run is False:
                continue
            if should_run is None:
                # only if one of the subsequent steps should run
                for subsequent in steps[stepidx + 1 :]:
                    subsequent_should_run = subsequent.should_run()
                    if subsequent_should_run is not None:
                        should_run = subsequent_should_run
                        break
                if should_run is False:
                    continue
            step.run(data)
            data.save()

    def get_steps(self):
        return self.steps

    @abc.abstractmethod
    def should_run(self):
        if self.steps:
            steps = self.get_steps()
            for step in steps:
                if isinstance(step, str):
                    getattr(self, step)()
                should_run = step.should_run()
                if should_run:
                    return True
            return False


class WizardStep(WizardBase):
    widgets = ()
    validate_functions = ()
    heading = ""
    pause = None

    def __init__(
        self,
        *widgets,
        validate=None,
        heading=None,
        pause=False,
        steps=None,
        **kwargs,
    ):
        super().__init__(steps)
        self.widgets = tuple(widgets or self.widgets)
        if not validate:
            validate = []
        elif not isinstance(validate, (list, tuple)):
            validate = tuple(validate)
        self.validate_functions = tuple(validate or self.validate_functions)
        self.heading = heading or self.heading
        self.pause = pause or self.pause

    def run(self, data: Config):
        self.data = data
        self.on_before_heading()
        if self.heading and not data.silent:
            heading = self.heading
            if callable(heading):
                heading = heading(data)
            if heading:
                slow_print(f"\n\n{Fore.BLUE}{heading}{Style.RESET_ALL}")
            print()
        self.on_after_heading()
        valid = False
        while not valid:
            widgets = self.get_widgets()
            for widget in widgets:
                widget_value = data.get(widget.name)
                if widget_value is None and widget.default:
                    if callable(widget.default):
                        widget_value = widget.default(data)
                    else:
                        widget_value = copy.deepcopy(widget.default)
                if data.no_input:
                    if widget_value is None:
                        raise ValueError(
                            f"Do not have a value for option {widget.name}, please specify it in the config"
                        )
                    if widget.value != widget_value:
                        data[widget.name] = widget_value
                    # propagate default if needed
                    continue
                widget.value = widget_value
                data[widget.name] = widget.run()
            valid = True
            validate_functions = self.get_validate_functions()
            for v in validate_functions:
                res = v(data)
                if res is True or res is None:
                    continue
                print(f"{Fore.RED}Error: {res}{Style.RESET_ALL}")
                valid = False
            if not valid and data.no_input:
                raise ValueError(f"Config not valid and not running interactively")
        super().run(data)
        if self.pause and not data.no_input:
            input(f"Press enter to continue ...")
        self.after_run()

    def on_before_heading(self):
        pass

    def on_after_heading(self):
        pass

    def get_widgets(self):
        return self.widgets

    def get_validate_functions(self):
        return self.validate_functions

    def on_after_steps(self):
        pass

    def after_run(self):
        pass


class InputWizardStep(WizardStep):
    def __init__(
        self,
        key,
        heading=None,
        required=True,
        default=None,
        prompt=None,
        force_run=False,
    ):
        super().__init__(
            Input(key, default=default, prompt=prompt),
            heading=heading,
            validate=[required_validation(key)] if required else [],
        )
        self.key = key
        self.force_run = force_run

    def should_run(self):
        return self.force_run or self.key not in self.data


class StaticWizardStep(WizardStep):
    def __init__(self, heading, **kwargs):
        super().__init__(heading=heading, **kwargs)

    def should_run(self):
        # do not know - should run only if one of the subsequent steps should run
        return None


class RadioWizardStep(WizardStep):
    def __init__(self, key, heading=None, options=None, default=None, force_run=False):
        super().__init__(Radio(key, default=default, options=options), heading=heading)
        self.key = key
        self.force_run = force_run

    def should_run(self):
        return self.force_run or self.key not in self.data



import re
import subprocess

from oarepo_cli.cli.site.utils import SiteWizardStepMixin
from oarepo_cli.ui.radio import Radio
from oarepo_cli.ui.wizard import WizardStep

from ...utils import run_cmdline


class InitDatabaseStep(SiteWizardStepMixin, WizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            Radio(
                "init_database",
                options={"yes": "Yes", "no": "No"},
                default="yes",
            ),
            heading="""
If the database has not been initialized, I can do it now - 
this will delete all the previous data in the database.

Should I do it?
            """,
            **kwargs,
        )

    def after_run(self):
        if self.data["init_database"] == "yes":
            run_cmdline(
                "pipenv",
                "run",
                "invenio",
                "db",
                "create",
                cwd=self.site_dir,
                environ={"PIPENV_IGNORE_VIRTUALENVS": "1"},
            )
        self.check_db_initialized(raise_error=True)

    def check_db_initialized(self, raise_error=False):
        try:
            output = run_cmdline(
                "pipenv",
                "run",
                "invenio",
                "alembic",
                "current",
                cwd=self.site_dir,
                environ={"PIPENV_IGNORE_VIRTUALENVS": "1"},
                grab_stdout=True,
                raise_exception=True,
            )
        except subprocess.CalledProcessError:
            raise Exception(
                "Alembic initialization failed. This could mean that the database "
                "does not exist or that there is already an incompatible alembic "
                "(previous repository) present in the database. Please fix the problem "
                "and try again"
            )
        if re.search("[a-zA-Z0-9]{12,24}\s+->\s+[a-zA-Z0-9]{12,24}", output):
            return True
        if raise_error:
            raise Exception(
                f"DB initialization failed. Expected alembic heads, got\n{output}"
            )
        return False

    def should_run(self):
        return not self.check_db_initialized(raise_error=False)



import click as click
from pathlib import Path
import yaml
import subprocess
from distutils.dir_util import copy_tree
import os
import errno
import time
import traceback
import sys
import select
import shutil


@click.command(
    name="docker-develop",
    hidden=True,
    help="Internal action called inside the development docker. "
    "Do not call from outside as it will not work. "
    "Call from the directory containing the oarepo.yaml",
)
@click.option("--virtualenv", help="Path to invenio virtualenv")
@click.option("--invenio", help="Path to invenio instance directory")
def docker_develop(**kwargs):
    call_task(install_editable_sources, **kwargs)
    call_task(db_init, **kwargs)
    call_task(search_init, **kwargs)
    call_task(create_custom_fields, **kwargs)
    call_task(import_fixtures, **kwargs)
    call_task(build_assets, **kwargs)
    call_task(development_script, **kwargs)

    runner = Runner(kwargs["virtualenv"], kwargs["invenio"])
    runner.run()


def call_task(task_func, **kwargs):
    status_file = Path(kwargs["invenio"]) / "docker-develop.yaml"
    if status_file.exists():
        with open(status_file, "r") as f:
            status = yaml.safe_load(f)
    else:
        status = {}
    if task_func.__name__ in status:
        return
    print(f"Calling task {task_func.__name__} with arguments {kwargs}")
    task_func(**kwargs)
    status[task_func.__name__] = True
    with open(status_file, "w") as f:
        yaml.safe_dump(status, f)


def install_editable_sources(*, virtualenv, **kwargs):
    """
    Editable sources are stored at virtualenv/requirements-editable.txt
    """
    check_call(
        [
            f"{virtualenv}/bin/pip",
            "install",
            "--no-deps",  # do not install dependencies as they were installed during container build
            "-r",
            f"{virtualenv}/requirements-editable.txt",
        ]
    )


def db_init(*, virtualenv, **kwargs):
    """
    Create database tables.
    """
    call([f"{virtualenv}/bin/invenio", "db", "drop", "--yes-i-know"])
    check_call([f"{virtualenv}/bin/invenio", "db", "create"])


def search_init(*, virtualenv, **kwargs):
    """
    Create search indices.
    """
    call([f"{virtualenv}/bin/invenio", "index", "destroy", "--force", "--yes-i-know"])
    check_call([f"{virtualenv}/bin/invenio", "index", "init"])


def create_custom_fields(*, virtualenv, **kwargs):
    """
    Create custom fields and patch indices.
    """
    check_call([f"{virtualenv}/bin/invenio", "oarepo", "cf", "init"])


def import_fixtures(*, virtualenv, **kwargs):
    """
    Import fixtures.
    """
    check_call([f"{virtualenv}/bin/invenio", "oarepo", "fixtures", "load"])


def development_script(**kwargs):
    if Path("development/initialize.sh").exists():
        check_call(["/bin/sh", "development/initialize.sh"])


# region Taken from Invenio-cli
#
# this and the following were taken from:
# https://github.com/inveniosoftware/invenio-cli/blob/0a49d438dc3c5ace872ce27f8555b401c5afc6e7/invenio_cli/commands/local.py#L45
# and must be called from the site directory
#
# The reason is that symlinking functionality is only part of invenio-cli
# and that is dependent on pipenv, which can not be used inside alpine
# (because we want to keep the image as small as possible, we do not install gcc
# and can only use compiled native python packages - like cairocffi or uwsgi). The
# version of these provided in alpine is slightly lower than the one created by Pipenv
# - that's why we use plain invenio & pip here.
#
# Another reason is that invenio-cli is inherently unstable when non-rdm version
# is used - it gets broken with each release.


def build_assets(*, virtualenv, invenio, **kwargs):
    shutil.rmtree(f"{invenio}/assets", ignore_errors=True)
    shutil.rmtree(f"{invenio}/static", ignore_errors=True)

    Path(f"{invenio}/assets").mkdir(parents=True)
    Path(f"{invenio}/static").mkdir(parents=True)

    check_call([f"{virtualenv}/bin/invenio", "collect", "--verbose"])
    check_call([f"{virtualenv}/bin/invenio", "webpack", "clean", "create"])
    check_call([f"{virtualenv}/bin/invenio", "webpack", "install"])

    copied_files = copy_statics_and_assets(invenio)
    symlink_assets_templates(copied_files, invenio)

    check_call([f"{virtualenv}/bin/invenio", "webpack", "build"])

    # do not allow Clean plugin to remove files
    webpack_config = Path(f"{invenio}/assets/build/webpack.config.js").read_text()
    webpack_config = webpack_config.replace("dry: false", "dry: true")
    Path(f"{invenio}/assets/build/webpack.config.js").write_text(webpack_config)


def copy_statics_and_assets(invenio):
    rdm_static_dir_exists = os.path.exists("static")
    rdm_assets_dir_exists = os.path.exists("assets")

    if rdm_static_dir_exists:
        src_dir = "static"
        dst_dir = f"{invenio}/static"
        for i in range(5):
            try:
                copy_tree(src_dir, dst_dir)
                break
            except:
                traceback.print_exc()
                time.sleep(10)
        else:
            raise Exception("Could not copy tree, see the log above")

    if rdm_assets_dir_exists:
        src_dir = "assets"
        dst_dir = f"{invenio}/assets"
        # The full path to the files that were copied is returned
        for i in range(5):
            try:
                ret = copy_tree(src_dir, dst_dir)
                break
            except:
                traceback.print_exc()
                time.sleep(10)
        else:
            raise Exception("Could not copy tree, see log above")
        return ret
    return []


def symlink_assets_templates(files_to_link, invenio):
    """Symlink the assets folder."""
    assets = "assets"
    click.secho("Symlinking {}...".format(assets), fg="green")

    instance_path = Path(invenio)
    project_dir = Path.cwd()
    for file_path in files_to_link:
        file_path = Path(file_path)
        relative_path = file_path.relative_to(instance_path)
        target_path = project_dir / relative_path
        force_symlink(target_path, file_path)


def force_symlink(target, link_name):
    """Forcefully create symlink at link_name pointing to target."""
    output = f"Symlinked {target} successfully."
    try:
        os.symlink(target, link_name)
    except OSError as e:
        if e.errno == errno.EEXIST:
            os.remove(link_name)
            os.symlink(target, link_name)
            output = output + "Deleted already existing link."

    print(target, output)


class Runner:
    def __init__(self, venv, invenio):
        self.venv = venv
        self.invenio = invenio
        self.server_handle = None
        self.ui_handle = None

    def run(self):
        try:
            self.start_server()
            time.sleep(10)
            self.start_ui()
        except:
            traceback.print_exc()
            self.stop_server()
            self.stop_ui()
            return

        while True:
            try:
                l = input_with_timeout(60)
                if not l:
                    continue
                if l == "stop":
                    break
                if l == "server":
                    self.stop_server()
                    self.start_server()
                    continue
                if l == "ui":
                    self.stop_ui()
                    self.start_ui()
                    continue
                if l == "build":
                    self.stop_ui()
                    self.stop_server()
                    build_assets(virtualenv=self.venv, invenio=self.invenio)
                    self.start_server()
                    time.sleep(10)
                    self.start_ui()

            except InterruptedError:
                self.stop_server()
                self.stop_ui()
                return
            except Exception:
                traceback.print_exc()
        self.stop_server()
        self.stop_ui()

    def start_server(self):
        print("Starting server")
        self.server_handle = subprocess.Popen(
            [
                f"{self.venv}/bin/invenio",
                "run",
                "--cert",
                "docker/nginx/test.crt",
                "--key",
                "docker/nginx/test.key",
                "-h",
                "0.0.0.0",
                "-p",
                "5000",
            ],
            env={
                "INVENIO_TEMPLATES_AUTO_RELOAD": "1",
                "FLASK_DEBUG": "1",
                **os.environ,
            },
            stdin=subprocess.DEVNULL,
        )

    def stop_server(self):
        print("Stopping server")
        self.stop(self.server_handle)
        self.server_handle = None

    def stop(self, handle):
        if handle:
            try:
                handle.terminate()
            except:
                pass
            time.sleep(5)
            try:
                handle.kill()
            except:
                pass

    def start_ui(self):
        print("Starting ui watcher")
        self.ui_handle = subprocess.Popen(
            ["npm", "run", "start"], cwd=f"{self.invenio}/assets"
        )

    def stop_ui(self):
        print("Stopping ui watcher")
        self.stop(self.ui_handle)
        self.ui_handle = None


#
# end of code taken from invenio-cli
# endregion


def check_call(*args, **kwargs):
    cmdline = " ".join(args[0])
    print(f"Calling command {cmdline} with kwargs {kwargs}")
    return subprocess.check_call(*args, **kwargs)


def call(*args, **kwargs):
    cmdline = " ".join(args[0])
    print(f"Calling command {cmdline} with kwargs {kwargs}")
    return subprocess.call(*args, **kwargs)


def input_with_timeout(timeout):
    print("=======================================================================")
    print()
    print("Type: ")
    print()
    print("    server <enter>    --- restart server")
    print("    ui <enter>        --- restart ui watcher")
    print("    build <enter>     --- stop server and watcher, ")
    print("                          call ui build, then start again")
    print("    stop <enter>      --- stop the server and ui and exit")
    print()
    i, o, e = select.select([sys.stdin], [], [], timeout)

    if i:
        return sys.stdin.readline().strip()



import click as click

from oarepo_cli.cli.utils import with_config


@click.command(name="list", help="List installed user interfaces")
@with_config()
def list_uis(cfg=None, **kwargs):
    for ui in cfg.whole_data.get("ui", {}):
        print(ui)



from oarepo_cli.ui.radio import Radio
from oarepo_cli.ui.wizard import WizardStep
from oarepo_cli.ui.wizard.validation import required


class DeploymentTypeStep(WizardStep):
    def __init__(self):
        super().__init__(
            Radio(
                name="packaging",
                options={
                    "single": "Build a single package out of all sources",
                    "multiple": "Build a separate package for each part of repository",
                },
                default="single",
            ),
            heading="""
Build a single python package from sources ("deployment" monorepo) or multiple packages?
Single package is the preferred choice unless you plan to share parts of repository
(such as metadata models, ui, ...) with other projects.            
            """,
            validate=[required("packaging")],
        )

    def should_run(self):
        return self.data.get("packaging") not in ("single", "multiple")



import os
import re
import subprocess
import sys
from pathlib import Path

import git
import pydriller
import pyfiglet
import tomlkit
from colorama import Fore, Style

from oarepo_cli.ui.utils import slow_print


def print_banner():
    intro_string = "\n".join(
        [
            "        " + x
            for x in pyfiglet.figlet_format(
                "O A R e p o", font="slant", width=120
            ).split("\n")
        ]
    )
    slow_print(f"\n\n\n{Fore.GREEN}{intro_string}{Style.RESET_ALL}")


def run_cmdline(
    *cmdline,
    cwd=".",
    environ=None,
    check_only=False,
    grab_stdout=False,
    grab_stderr=False,
    discard_output=False,
    raise_exception=False,
):
    env = os.environ.copy()
    env.update(environ or {})
    cwd = Path(cwd).absolute()
    cmdline = [str(x) for x in cmdline]
    print(
        f"{Fore.BLUE}Running {Style.RESET_ALL} {' '.join(cmdline)}", file=sys.__stderr__
    )
    print(f"{Fore.BLUE}    inside {Style.RESET_ALL} {cwd}", file=sys.__stderr__)
    try:
        if grab_stdout or grab_stderr or discard_output:
            kwargs = {}
            if grab_stdout or discard_output:
                kwargs["stdout"] = subprocess.PIPE
            if grab_stderr or discard_output:
                kwargs["stderr"] = subprocess.PIPE

            ret = subprocess.run(
                cmdline,
                check=True,
                cwd=cwd,
                env=env,
                **kwargs,
            )
            ret = (ret.stdout or b"") + b"\n" + (ret.stderr or b"")
        else:
            ret = subprocess.call(cmdline, cwd=cwd, env=env)
            if ret:
                raise subprocess.CalledProcessError(ret, cmdline)
    except subprocess.CalledProcessError as e:
        if check_only:
            return False
        print(f"Error running {' '.join(cmdline)}", file=sys.__stderr__)
        if e.stdout:
            print(e.stdout)
        if e.stderr:
            print(e.stderr)
        if raise_exception:
            raise
        sys.exit(e.returncode)
    print(
        f"{Fore.GREEN}Finished running {Style.RESET_ALL} {' '.join(cmdline)}",
        file=sys.__stderr__,
    )
    print(f"{Fore.GREEN}    inside {Style.RESET_ALL} {cwd}", file=sys.__stderr__)
    if grab_stdout:
        return ret.decode("utf-8").strip()
    return True


def find_oarepo_project(dirname, raises=False):
    dirname = Path(dirname).absolute()
    orig_dirname = dirname
    for _ in range(4):
        if (dirname / "oarepo.yaml").exists():
            return dirname
        dirname = dirname.parent
    if raises:
        raise Exception(
            f"Not part of OARepo project: directory {orig_dirname} "
            f"or its 4 ancestors do not contain oarepo.yaml file"
        )
    return


def add_to_pipfile_dependencies(pipfile, package_name, package_path):
    with open(pipfile, "r") as f:
        pipfile_data = tomlkit.load(f)
    for pkg in pipfile_data["packages"]:
        if pkg == package_name:
            break
    else:
        t = tomlkit.inline_table()
        t.comment("Added by nrp-cli")
        t.update({"editable": True, "path": package_path})
        pipfile_data["packages"].add(tomlkit.nl())
        pipfile_data["packages"][package_name] = t
        pipfile_data["packages"].add(tomlkit.nl())

        with open(pipfile, "w") as f:
            tomlkit.dump(pipfile_data, f)


def to_python_name(x):
    x = re.sub(r"(?<!^)(?=[A-Z])", "_", x).lower()
    x = x.replace("-", "_")
    return re.sub("[^a-z_]", "", x)


def pip_install(pip_binary, env_name, lib_name_and_version, lib_github):
    # run pip installation, taking env versions into account
    run_cmdline(pip_binary, "install", "-U", "--no-input", "setuptools", "pip", "wheel")
    installation_option = os.environ.get(env_name, "release")
    if installation_option == "release":
        # release
        run_cmdline(pip_binary, "install", "--no-input", lib_name_and_version)
    elif installation_option == "maintrunk":
        run_cmdline(
            pip_binary,
            "install",
            "--no-input",
            f"git+{lib_github}",
        )
    elif installation_option.startswith("https://"):
        run_cmdline(
            pip_binary,
            "install",
            "--no-input",
            f"git+{installation_option}",
        )
    else:
        run_cmdline(
            pip_binary,
            "install",
            "--no-input",
            "-e",
            Path(installation_option),
        )


def get_cookiecutter_source(env_name, lib_github, lib_version, master_version="master"):
    installation_option = os.environ.get(env_name, "release")
    if installation_option == "release":
        cookiecutter_path = lib_github
        cookiecutter_branch = lib_version
    elif installation_option == "maintrunk":
        cookiecutter_path = lib_github
        cookiecutter_branch = master_version
    elif installation_option.startswith("https://"):
        # something like https://github.com/oarepo/oarepo-model-builder/tree/datatypes
        cookiecutter_path, cookiecutter_branch = installation_option.rsplit(
            "/tree/", maxsplit=1
        )
    else:
        cookiecutter_path = installation_option
        cookiecutter_branch = None
    return cookiecutter_path, cookiecutter_branch


def commit_git(repo_dir, tag_name, message):
    tag_index = 1
    try:
        for commit in pydriller.Repository(str(repo_dir)).traverse_commits():
            for m in commit.msg.split("\n"):
                if m.startswith("omb-"):
                    tag_index += 1
    except git.exc.GitCommandError:
        pass
    repo = git.Repo(repo_dir)
    tag_name = f"omb-{tag_index:05d}-{tag_name}"
    index = repo.index
    repo.git.add(repo_dir)
    index.commit(message + "\n\n" + tag_name)


def must_be_committed(repo_dir):
    repo = git.Repo(repo_dir)
    if repo.is_dirty() or repo.untracked_files:
        for f in repo.untracked_files:
            print("    ", f)

        print(
            "The repository contains untracked or dirty files. Please commit/ignore them before continuing."
        )
        sys.exit(1)



import functools
import shutil
import sys
from pathlib import Path

import click
import yaml
from colorama import Fore, Style

from oarepo_cli.config import MonorepoConfig
from oarepo_cli.ui.wizard.steps import RadioWizardStep, WizardStep
from oarepo_cli.utils import add_to_pipfile_dependencies, run_cmdline


def with_config(
    config_section=None, project_dir_as_argument=False, config_as_argument=False
):
    def wrapper(f):
        @(
            click.argument(
                "project_dir",
                type=click.Path(exists=False, file_okay=False),
                required=True,
            )
            if project_dir_as_argument
            else click.option(
                "--project-dir",
                type=click.Path(exists=True, file_okay=False),
                required=False,
                help="Directory containing an already initialized project. "
                "If not set, current directory is used",
            )
        )
        @click.option(
            "--no-banner",
            is_flag=True,
            type=bool,
            required=False,
            help="Do not show the welcome banner",
        )
        @click.option(
            "--no-input",
            is_flag=True,
            type=bool,
            required=False,
            help="Take options from the config file, skip user input",
        )
        @click.option(
            "--silent",
            is_flag=True,
            type=bool,
            required=False,
            help="Do not output program's messages. "
            "External program messages will still be displayed",
        )
        @click.option(
            "--verbose",
            is_flag=True,
            type=bool,
            required=False,
            help="Verbose output",
        )
        @(
            click.argument(
                "config",
                type=click.Path(exists=True, file_okay=True, dir_okay=False),
                required=True,
            )
            if config_as_argument
            else click.option(
                "--config",
                type=click.Path(exists=True, file_okay=True, dir_okay=False),
                required=False,
                help="Merge this config to the main config in target directory and proceed",
            )
        )
        @functools.wraps(f)
        def wrapped(
            project_dir=None,
            no_input=False,
            silent=False,
            config=None,
            verbose=False,
            **kwargs,
        ):
            if not project_dir:
                project_dir = Path.cwd()
            project_dir = Path(project_dir).absolute()
            oarepo_yaml_file = project_dir / "oarepo.yaml"

            if callable(config_section):
                section = config_section(**kwargs)
            else:
                section = config_section or "config"

            cfg = MonorepoConfig(oarepo_yaml_file, section=section)

            if oarepo_yaml_file.exists():
                cfg.load()

            if config:
                config_data = yaml.safe_load(Path(config).read_text())
                cfg.merge_config(config_data, top=not config_section)

            cfg.no_input = no_input
            cfg.silent = silent
            cfg.verbose = verbose

            kwargs.pop("cfg", None)
            kwargs.pop("project_dir", None)
            try:
                return f(project_dir=project_dir, cfg=cfg, **kwargs)
            except Exception as e:
                if cfg.verbose:
                    import traceback

                    traceback.print_exc()
                else:
                    print(str(e))
                raise
                sys.exit(1)

        return wrapped

    return wrapper


class ProjectWizardMixin:
    @property
    def site_dir(self):
        if not hasattr(self, "site"):
            raise Exception("Current site not set")
        return self.data.project_dir / self.site["site_dir"]

    @property
    def invenio_cli(self):
        return self.data.project_dir / self.data.get("config.invenio_cli")

    @property
    def oarepo_cli(self):
        return self.data.project_dir / self.data.get("config.oarepo_cli")

    def invenio_cli_command(self, *args, cwd=None, environ=None):
        return run_cmdline(
            self.invenio_cli,
            *args,
            cwd=cwd or self.site_dir,
            environ={"PIPENV_IGNORE_VIRTUALENVS": "1", **(environ or {})},
        )

    def pipenv_command(self, *args, cwd=None, environ=None):
        return run_cmdline(
            "pipenv",
            *args,
            cwd=cwd or self.site_dir,
            environ={"PIPENV_IGNORE_VIRTUALENVS": "1", **(environ or {})},
        )

    def invenio_command(self, *args, cwd=None, environ=None):
        return run_cmdline(
            "pipenv",
            "run",
            "invenio",
            *args,
            cwd=cwd or self.site_dir,
            environ={"PIPENV_IGNORE_VIRTUALENVS": "1", **(environ or {})},
        )

    def run_cookiecutter(
        self,
        template,
        config_file,
        checkout=None,
        output_dir=None,
        extra_context=None,
        environ=None,
    ):
        config_dir: Path = self.data.project_dir / ".cookiecutters"
        config_dir.mkdir(parents=True, exist_ok=True)
        config_file = config_dir / f"{config_file}.yaml"
        config_file.write_text(yaml.safe_dump({"default_context": extra_context}))
        cookiecutter_command = (
            Path(sys.executable or sys.argv[0]).absolute().parent / "cookiecutter"
        )
        output_dir_temp = f"{output_dir}-tmp"
        output_dir = Path(output_dir)
        output_dir_temp = Path(output_dir_temp)
        args = [
            template,
            "--no-input",
            "-o",
            output_dir_temp,
            "--config-file",
            config_file,
        ]
        if checkout:
            args.append("-c")
            args.append(checkout)

        run_cmdline(
            cookiecutter_command,
            *args,
            cwd=self.data.project_dir,
            environ={"PIPENV_IGNORE_VIRTUALENVS": "1", **(environ or {})},
        )
        merge_from_temp_to_target(output_dir_temp, output_dir)


class SiteMixin(ProjectWizardMixin):
    @property
    def site_dir(self):
        site_name = self.data.get("installation_site", None)
        if not site_name:
            raise Exception("Unexpected error: No installation site specified")
        site = self.data.get(f"sites.{site_name}")
        if not site:
            raise Exception(
                f"Unexpected error: Site with name {site_name} does not exist"
            )
        return self.data.project_dir / site["site_dir"]


class PipenvInstallWizardStep(SiteMixin, ProjectWizardMixin, WizardStep):
    folder = None

    def get_steps(self):
        sites = self.data.whole_data.get("sites", {})
        if len(sites) == 1:
            self.data["installation_site"] = next(iter(sites))
            steps = []
        else:
            steps = [
                RadioWizardStep(
                    "installation_site",
                    options={
                        x: f"{Fore.GREEN}{x}{Style.RESET_ALL}"
                        for x in self.data.whole_data["sites"]
                    },
                    default=next(iter(self.data.whole_data["sites"])),
                    heading=f"""
            Select the site where you want to install the model to.
                """,
                    force_run=True,
                )
            ]
        return steps

    def heading(self, data):
        return f"""
    Now I will add the {self.folder} to site's Pipfile (if it is not there yet)
    and will run pipenv lock & install.
        """

    pause = True

    def after_run(self):
        # add package to pipfile
        self.add_to_pipfile()
        self.install_pipfile()

    def add_to_pipfile(self):
        pipfile = self.site_dir / "Pipfile"
        add_to_pipfile_dependencies(
            pipfile, self.data.section, f"../../{self.folder}/{self.data.section}"
        )

    def install_pipfile(self):
        self.pipenv_command("lock")
        self.pipenv_command("install")

    def should_run(self):
        return True


def merge_from_temp_to_target(output_dir_temp, output_dir):
    source: Path
    for source in output_dir_temp.rglob("*"):
        rel = source.relative_to(output_dir_temp)
        dest: Path = output_dir / rel
        if source.is_dir():
            if not dest.exists():
                dest.mkdir(parents=True)
        elif source.is_file():
            if not dest.exists():
                if not dest.parent.exists():
                    dest.parent.mkdir(parents=True)
                shutil.copy(source, dest)
    shutil.rmtree(output_dir_temp)



from __future__ import annotations

import datetime
import os
import re

from oarepo_cli.cli.site.utils import SiteWizardStepMixin
from oarepo_cli.ui.wizard import StaticWizardStep, WizardStep
from oarepo_cli.ui.wizard.steps import InputWizardStep, RadioWizardStep

from ...utils import commit_git, get_cookiecutter_source, run_cmdline


class InstallSiteStep(SiteWizardStepMixin, WizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            heading="""
Now I will add site sources, that can be used to change the overall CSS style and
configure your repository. Please fill in the following values. 
If not sure, keep the default values.""",
            **kwargs,
        )

    def get_steps(self):
        self.data.setdefault(
            "transifex_project", self.data.get("config.project_package", "")
        )
        # substeps of this step
        site_dir_name = self.site_dir.name
        return [
            InputWizardStep(
                "repository_name",
                prompt="""Enter the repository name ("title" of the HTML site)""",
                default=re.sub("[_-]", " ", site_dir_name).title(),
            ),
            InputWizardStep(
                "www",
                prompt="""Enter the WWW address on which the repository will reside""",
            ),
            InputWizardStep(
                "author_name",
                prompt="""Author name""",
                default=os.environ.get("USERNAME") or os.environ.get("USER"),
            ),
            InputWizardStep("author_email", prompt="""Author email"""),
            InputWizardStep(
                "year",
                prompt="""Year (for copyright)""",
                default=datetime.datetime.today().strftime("%Y"),
            ),
            InputWizardStep("copyright_holder", prompt="""Copyright holder"""),
            RadioWizardStep(
                "use_oarepo_vocabularies",
                options={"yes": "Yes", "no": "No"},
                default="yes",
                heading=f"""
            Are you planning to use extended vocabularies (extra fields on vocabularies, hierarchy in vocabulary items)? If in doubt, select 'yes'.
                """,
            ),
            StaticWizardStep(
                heading="""I have all the information to generate the site.
To do so, I'll call the invenio client. If anything goes wrong, please fix the problem
and run the wizard again.
            """,
            ),
        ]

    def after_run(self):
        # create site config for invenio-cli
        cookiecutter_config_file = self.data.project_dir / ".invenio"
        site_dir = self.site_dir
        if not site_dir.parent.exists():
            site_dir.parent.mkdir(parents=True)

        with open(cookiecutter_config_file, "w") as f:
            print(
                f"""
[cookiecutter]
project_name = {self.data['repository_name']}
project_shortname = {self.site_dir.name}
project_site = {self.data['www']}
github_repo = 
description = {self.data['repository_name']} OARepo Instance
author_name = {self.data['author_name']}
author_email = {self.data['author_email']}
year = {self.data['year']}
python_version = 3.9
database = postgresql
search = opensearch2
file_storage = S3
development_tools = yes
site_code = yes
use_oarepo_vocabularies = {self.data['use_oarepo_vocabularies']}
                """,
                file=f,
            )
        # and run invenio-cli with our site template
        # (submodule from https://github.com/oarepo/cookiecutter-oarepo-instance)

        cookiecutter_path, cookiecutter_branch = get_cookiecutter_source(
            "OAREPO_SITE_COOKIECUTTER_VERSION",
            "https://github.com/oarepo/cookiecutter-site",
            "v11.0",
            master_version="master",
        )

        run_cmdline(
            self.data.project_dir / self.data.get("config.invenio_cli"),
            "init",
            "rdm",
            "-t",
            cookiecutter_path,
            *(
                [
                    "-c",
                    cookiecutter_branch,
                ]
                if cookiecutter_branch
                else []
            ),
            "--no-input",
            "--config",
            str(cookiecutter_config_file),
            cwd=self.data.project_dir / "sites",
            environ={
                "PIPENV_IGNORE_VIRTUALENVS": "1",
                # use our own cookiecutter, not the system one
                "PATH": f"{self.data.get('config.project_dir')}/.bin:{os.environ['PATH']}",
            },
        )
        with open(self.site_dir / ".check.ok", "w") as f:
            f.write("oarepo check ok")
        commit_git(
            self.data.project_dir,
            f"after-site-cookiecutter-{self.data.section}",
            f"Committed automatically after site {self.data.section} cookiecutter has been called",
        )

    def should_run(self):
        return not (self.site_dir / ".check.ok").exists()



import json
from os.path import relpath
from pathlib import Path
import re
from typing import Any, List

import click as click

from oarepo_cli.cli.model.utils import ProjectWizardMixin
from oarepo_cli.cli.utils import with_config
from oarepo_cli.ui.wizard import StaticWizardStep, Wizard
from oarepo_cli.ui.wizard.steps import InputWizardStep, RadioWizardStep, WizardStep
from oarepo_cli.utils import get_cookiecutter_source, run_cmdline, to_python_name
from ...utils import commit_git


@click.command(
    name="add",
    help="""Generate a new UI. Required arguments:
    <name>   ... name of the ui. The recommended pattern for it is <modelname>-ui
    """,
)
@click.argument("name")
@with_config(config_section=lambda name, **kwargs: ["ui", name])
def add_ui(cfg=None, **kwargs):
    commit_git(
        cfg.project_dir,
        f"before-ui-add-{cfg.section}",
        f"Committed automatically before ui {cfg.section} has been added",
    )
    add_ui_wizard(cfg).run(cfg)
    commit_git(
        cfg.project_dir,
        f"after-ui-add-{cfg.section}",
        f"Committed automatically after ui {cfg.section} has been added",
    )


class UIWizardMixin:
    @property
    def ui_name(self):
        return self.data.section

    @property
    def ui_dir(self):
        return self.data.project_dir / "ui" / self.ui_name


def available_models(data):
    known_models = {
        # TODO: model description while adding models
        x: x
        for x in data.whole_data.get("models", {}).keys()
    }
    return known_models


def snail_to_title(v):
    return "".join(ele.title() for ele in v.split("_"))


class ModelMixin:
    def get_model_definition(self):
        model_config = self.data.whole_data["models"][self.data["model_name"]]
        model_package = model_config["model_package"]

        model_path = self.data.project_dir / model_config["model_dir"]
        model_file = (
            (model_path / model_package / "models" / "model.json")
            .absolute()
            .resolve(strict=False)
        )
        with open(model_file) as f:
            model_description = json.load(f)
        return model_description, model_path, model_package, model_config


class AddUIWizardStep(ModelMixin, UIWizardMixin, ProjectWizardMixin, WizardStep):
    def after_run(self):
        ui_name = self.ui_name

        ui_package = to_python_name(ui_name)
        ui_base = snail_to_title(ui_package)

        (
            model_description,
            model_path,
            model_package,
            model_config,
        ) = self.get_model_definition()

        model_service = model_description["model"]["service-id"]
        ui_serializer_class = model_description["model"]["record-ui-serializer-class"]

        self.data.setdefault(
            "cookiecutter_local_model_path", relpath(model_path, self.ui_dir)
        )
        self.data.setdefault("cookiecutter_model_package", model_package)
        self.data.setdefault("cookiecutter_app_name", ui_name)
        self.data.setdefault("cookiecutter_app_package", ui_package)
        self.data.setdefault("cookiecutter_ext_name", f"{ui_base}Extension")

        self.data.setdefault("cookiecutter_author", model_config.get("author_name", ""))
        self.data.setdefault(
            "cookiecutter_author_email", model_config.get("author_email", "")
        )
        self.data.setdefault("cookiecutter_repository_url", "")
        # TODO: take this dynamically from the running model's Ext so that
        # TODO: it does not have to be specified here
        self.data.setdefault("cookiecutter_resource", f"{ui_base}Resource")
        self.data.setdefault("cookiecutter_resource_config", f"{ui_base}ResourceConfig")
        self.data.setdefault("cookiecutter_api_service", model_service)
        self.data.setdefault(
            "cookiecutter_ui_record_serializer_class", ui_serializer_class
        )

        cookiecutter_data = {
            "model_name": self.data["model_name"],
            "local_model_path": self.data["cookiecutter_local_model_path"],
            "model_package": self.data["cookiecutter_model_package"],
            "app_name": self.data["cookiecutter_app_name"],
            "app_package": self.data["cookiecutter_app_package"],
            "ext_name": self.data["cookiecutter_ext_name"],
            "author": self.data["cookiecutter_author"],
            "author_email": self.data["cookiecutter_author_email"],
            "repository_url": self.data["cookiecutter_repository_url"],
            # TODO: take this dynamically from the running model's Ext so that
            # TODO: it does not have to be specified here
            "resource": self.data["cookiecutter_resource"],
            "resource_config": self.data["cookiecutter_resource_config"],
            "api_service": self.data["cookiecutter_api_service"],
            "ui_serializer_class": self.data["cookiecutter_ui_record_serializer_class"],
            "url_prefix": self.data["url_prefix"],
        }

        cookiecutter_path, cookiecutter_branch = get_cookiecutter_source(
            "OAREPO_UI_COOKIECUTTER_VERSION",
            "https://github.com/oarepo/cookiecutter-app",
            "v11.0",
            master_version="master",
        )

        self.run_cookiecutter(
            template=cookiecutter_path,
            config_file=f"ui-{ui_name}",
            checkout=cookiecutter_branch,
            output_dir=self.data.project_dir / "ui",
            extra_context=cookiecutter_data,
        )
        self.data["ui_dir"] = f"ui/{ui_name}"

    def should_run(self):
        return not self.ui_dir.exists()


def add_ui_wizard(data):
    available = available_models(data)
    return Wizard(
        StaticWizardStep(
            heading="""
A UI is a python package that displays the search, detail, edit, ... pages for a single
metadata model. At first you'll have to select the model for which the UI will be created
and then I'll ask you a couple of additional questions.
""",
        ),
        AddUIWizardStep(
            steps=[
                RadioWizardStep(
                    "model_name",
                    heading="""
        For which model do you want to generate the ui?
        """,
                    options=available,
                    default=next(iter(available)),
                ),
                InputWizardStep(
                    "url_prefix",
                    prompt="On which url prefix will the UI reside? The prefix should like /something/: ",
                    default=lambda data: f"/{data.section}/",
                ),
            ]
        ),
        CreateJinjaStep(),
    )


class CreateJinjaStep(ModelMixin, WizardStep):
    def should_run(self):
        return True

    def after_run(self):
        (
            model_description,
            model_path,
            model_package,
            model_config,
        ) = self.get_model_definition()
        # get the UI definition
        ui_definition_path = model_path / model_package / "models" / "ui.json"
        ui_definition = json.loads(ui_definition_path.read_text())

        # get the first site (TODO: how to do this better?) and load registered renderers
        sites = self.data.whole_data.get("sites", [])
        if sites:
            first_site = next(iter(sites.values()))
            renderers_json = run_cmdline(
                "pipenv",
                "run",
                "invenio",
                "oarepo",
                "ui",
                "renderers",
                "--json",
                cwd=self.data.project_dir / first_site["site_dir"],
                environ={"PIPENV_IGNORE_VIRTUALENVS": "1"},
                grab_stdout=True,
            )
            renderers = [x["renderer"] for x in json.loads(renderers_json)]
        else:
            # pre-defined renderers
            renderers = [
                "array",
                "date",
                "datetime",
                "double",
                "field",
                "float",
                "fulltext",
                "fulltext__43__keyword",
                "int",
                "keyword",
                "time",
                "value",
            ]

        template, macro_definitions = self.generate_main(ui_definition)
        if macro_definitions:
            macros = "\n".join(
                self.generate_macro_definitions(macro_definitions, set(renderers))
            )
        else:
            macros = None

        # save template and macros
        ui_dir = self.data.project_dir / self.data.config["ui_dir"]
        main_jinja_path = (
            ui_dir
            / self.data.config["cookiecutter_app_package"]
            / "templates"
            / "semantic-ui"
            / self.data.config["cookiecutter_app_package"]
            / "main.html"
        )
        template = main_jinja_path.read_text() + "\n\n" + template
        main_jinja_path.write_text(template)

        macros_jinja_path: Path = (
            ui_dir
            / self.data.config["cookiecutter_app_package"]
            / "templates"
            / "semantic-ui"
            / "oarepo_ui"
            / "components"
            / "100-macros.html"
        )
        macros_jinja_path.parent.mkdir(exist_ok=True, parents=True)
        macros_jinja_path.write_text(macros or "")

    def _select(self, fields, *keys):
        for k in keys:
            if k in fields:
                return k, fields.pop(k)
        return None, None

    def generate_main(self, ui):
        macro_definitions = []
        template = []
        fields = ui["children"]
        if "metadata" in fields:
            md = fields.pop("metadata")
            fields.update({f"metadata.{k}": v for k, v in md["children"].items()})
        title_key, title = self._select(fields, "title", "metadata.title")
        divider = False
        if title_key:
            template.append(f'<h1>{{%- value "{title_key}" -%}}</h1>')
            macro_definitions.append(title)
            divider = True
        creator_key, creator = self._select(fields, "creator", "metadata.creator")
        if creator_key:
            template.append(
                f'<div class="creator">{{%- value "{creator_key}" -%}}</div>'
            )
            macro_definitions.append(creator)
            divider = True
        if divider:
            template.append('<hr class="divider"/>')
        template.append('<dl class="detail-fields">')
        for fld_key, fld in sorted(fields.items()):
            template.append(f'{{%- field "{fld_key}" -%}}')
            macro_definitions.append(fld)
        template.append("</dl>")

        return "\n".join(template), macro_definitions

    def generate_macro_definitions(
        self, macro_definitions: List[Any], processed_components
    ):
        for definition in macro_definitions:
            if not definition.get("detail"):
                continue
            component = re.sub(r"\W", replace_non_variable_signs, definition["detail"])

            if component in processed_components:
                _, children = self._generate_macro_children(definition)
            else:
                processed_components.add(component)

                children_def, children = self._generate_macro_children(definition)
                if children_def:
                    yield f"\n\n{{%- macro render_{component}(arg) -%}}\n<dl class='detail-subfields'>\n{children_def}\n</dl>\n{{%- endmacro -%}}"
                else:
                    yield f"\n\n{{%- macro render_{component}(arg) -%}}{'{{'}arg{'}}'}{{%- endmacro -%}}"

            yield from self.generate_macro_definitions(children, processed_components)

    def _generate_macro_children(self, definition):
        # for array members, do not return fields as array macro is built-in
        if "child" in definition:
            return "", [definition["child"]]
        if "children" not in definition:
            return "", []
        fields = []
        children = []
        for c_key, cdef in definition["children"].items():
            fields.append(f'{{%- field "{c_key}" -%}}')
            children.append(cdef)
        return "\n".join(fields), children


def replace_non_variable_signs(x):
    return f"__{ord(x.group())}__"



import click

from oarepo_cli.cli.utils import with_config

from ...ui.wizard import Wizard
from ...utils import commit_git, to_python_name
from .step_01_install_site import InstallSiteStep
from .step_02_check_requirements import CheckRequirementsStep
from .step_03_fixup_code import FixupSiteCodeStep
from .step_04_start_containers import StartContainersStep
from .step_05_create_pipenv import CreatePipenvStep
from .step_06_install_invenio import InstallInvenioStep
from .step_07_init_database import InitDatabaseStep
from .step_08_init_files import InitFilesStep
from .step_09_next_steps import NextStepsStep


@click.command(
    name="add",
    help="""Generate a new site.  Required arguments:
    <name>   ... name of the site. The recommended pattern for it is <something>-site""",
)
@click.argument("name")
@with_config(config_section=lambda name, **kwargs: ["sites", name])
@click.pass_context
def add_site(ctx, cfg=None, name=None, **kwargs):
    commit_git(
        cfg.project_dir,
        f"before-site-install-{cfg.section}",
        f"Committed automatically before site {cfg.section} has been added",
    )
    cfg["site_package"] = to_python_name(name)
    cfg["site_dir"] = f"sites/{name}"

    initialize_wizard = Wizard(
        InstallSiteStep(pause=True),
        CheckRequirementsStep(),
        FixupSiteCodeStep(),
        StartContainersStep(pause=True),
        CreatePipenvStep(),
        InstallInvenioStep(pause=True),
        InitDatabaseStep(),
        InitFilesStep(),
        NextStepsStep(pause=True),
    )
    initialize_wizard.run(cfg)
    commit_git(
        cfg.project_dir,
        f"after-site-install-{cfg.section}",
        f"Committed automatically after site {cfg.section} has been added",
    )



import os
import shutil
import venv
import re

import click as click
from colorama import Fore, Style
from setuptools.config import read_configuration

from oarepo_cli.cli.model.utils import ModelWizardStep
from oarepo_cli.cli.utils import PipenvInstallWizardStep, with_config
from oarepo_cli.ui.wizard import Wizard
from oarepo_cli.ui.wizard.steps import RadioWizardStep, WizardStep
from oarepo_cli.utils import commit_git, run_cmdline


@click.command(
    name="install",
    help="""
Install the model into the current site. Required arguments:
    <name>   ... name of the already existing model""",
)
@click.argument("name", required=True)
@with_config(config_section=lambda name, **kwargs: ["models", name])
def install_model(cfg=None, **kwargs):
    commit_git(
        cfg.project_dir,
        f"before-model-install-{cfg.section}",
        f"Committed automatically before model {cfg.section} has been installed",
    )
    wizard = Wizard(
        RadioWizardStep(
            "run_tests",
            options={
                "run": f"{Fore.GREEN}Run tests{Style.RESET_ALL}",
                "skip": f"{Fore.RED}Skip tests{Style.RESET_ALL}",
            },
            default="run",
            heading=f"""
        Before installing the model, it is wise to run the test to check that the model is ok.
        If the tests fail, please fix the errors and run this command again.
            """,
            force_run=True,
        ),
        TestWizardStep(),
        InstallWizardStep(),
        AlembicWizardStep(),
        UpdateIndexWizardStep(),
    )
    wizard.run(cfg)
    commit_git(
        cfg.project_dir,
        f"after-model-install-{cfg.section}",
        f"Committed automatically after model {cfg.section} has been installed",
    )


class TestWizardStep(ModelWizardStep, WizardStep):
    def after_run(self):
        if self.data["run_tests"] == "skip":
            return
        model_dir = self.model_dir
        venv_dir = model_dir / ".venv-test"
        if venv_dir.exists():
            shutil.rmtree(venv_dir)

        venv.main([str(venv_dir)])
        pip_binary = venv_dir / "bin" / "pip"
        pytest_binary = venv_dir / "bin" / "pytest"

        run_cmdline(
            pip_binary, "install", "-U", "--no-input", "setuptools", "pip", "wheel"
        )
        run_cmdline(
            pip_binary, "install", "--no-input", "-e", ".[tests]", cwd=model_dir
        )

        run_cmdline(
            pytest_binary,
            "tests",
            cwd=model_dir,
            environ={
                "OPENSEARCH_HOST": self.data.get(
                    "config.TEST_OPENSEARCH_HOST", "localhost"
                ),
                "OPENSEARCH_PORT": self.data.get("config.TEST_OPENSEARCH_PORT", "9400"),
            },
        )

    def should_run(self):
        return True


class InstallWizardStep(PipenvInstallWizardStep):
    folder = "models"


class AlembicWizardStep(ModelWizardStep):
    heading = f"""
    I will create/update the alembic migration steps so that you might later modify 
    the model and perform automatic database migrations. This command will write
    alembic steps (if the database layer has been modified) to the models' alembic directory.
                """
    pause = True

    def after_run(self):
        setup_cfg = read_configuration(self.model_dir / "setup.cfg")
        for alembic_config in setup_cfg["options"]["entry_points"]["invenio_db.models"]:
            branch, model_dir = [
                x.strip() for x in alembic_config.split("=", maxsplit=1)
            ]

            model_dir = model_dir.replace(".", os.sep)
            model_dir = self.model_dir / model_dir
            alembic_path = self.get_alembic_path(model_dir)
            if not alembic_path:
                print(
                    f"{Fore.RED}Alembic path not found for model in {model_dir}, "
                    f"so can not set up alembic. You will have to do it manually{Style.RESET_ALL}"
                )
            else:
                self.setup_alembic(branch, alembic_path)

    def get_alembic_path(self, model_dir):
        md = model_dir
        while md != self.model_dir:
            ap = md / "alembic"
            if ap.exists():
                return ap
            md = md.parent

    def setup_alembic(self, branch, alembic_path):
        filecount = len(list(alembic_path.iterdir()))
        revision_id_prefix = self.data['model_package']

        def rewrite_revision_file(file_suffix, new_id_number):
            files = list(alembic_path.iterdir())
            target_file = str([file_name for file_name in files if file_suffix in str(file_name)][0])
            id_start_index = target_file.rfind("/") + 1
            id_end_index = target_file.find(file_suffix)
            old_id = target_file[id_start_index:id_end_index]
            new_id = f"{revision_id_prefix}_{new_id_number}"
            with open(target_file, 'r') as f:
                file_text = f.read()
                file_text = file_text.replace(f"revision = '{old_id}'", f"revision = '{new_id}'")
            with open(target_file.replace(old_id, new_id), 'w') as f:
                f.write(file_text)
            os.remove(target_file)

        if filecount < 2:
            # alembic has not been initialized yet ...
            self.invenio_command("alembic", "upgrade", "heads")
            # create model branch
            self.invenio_command(
                "alembic",
                "revision",
                f"Create {branch} branch for {self.data['model_package']}.",
                "-b",
                branch,
                "-p",
                "dbdbc1b19cf2",
                "--empty",
            )


            rewrite_revision_file("_create_", "1")

            self.fix_sqlalchemy_utils(alembic_path)
            self.invenio_command("alembic", "upgrade", "heads")
            self.invenio_command(
                "alembic", "revision", "Initial revision.", "-b", branch
            )

            rewrite_revision_file("_initial_revision", "2") # the link to down-revision is created correctly after alembic upgrade heads on the corrected file, explicit rewrite of down-revision is not needed

            self.fix_sqlalchemy_utils(alembic_path)
            self.invenio_command("alembic", "upgrade", "heads")
        else:
            # alembic has been initialized, update heads and generate
            files = [file_path.name for file_path in alembic_path.iterdir()]

            file_numbers = []
            for file in files:
                file_number_regex = re.findall(f"(?<={revision_id_prefix}_)\d+", file)
                if file_number_regex:
                    file_numbers.append(int(file_number_regex[0]))
            new_file_number = max(file_numbers) + 1
            
            self.invenio_command("alembic", "upgrade", "heads")
            self.invenio_command(
                "alembic",
                "revision",
                "nrp-cli install revision.",
                "-b",
                branch,
            )

            rewrite_revision_file("_nrp_cli_install", new_file_number)

            self.fix_sqlalchemy_utils(alembic_path)
            self.invenio_command("alembic", "upgrade", "heads")

    def fix_sqlalchemy_utils(self, alembic_path):
        for fn in alembic_path.iterdir():
            data = fn.read_text()

            empty_migration = '''
def upgrade():
    """Upgrade database."""
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###'''

            if empty_migration in data:
                print(
                    f"{Fore.YELLOW}Found empty migration in file {fn}, deleting it{Style.RESET_ALL}"
                )
                fn.unlink()
                continue

            modified = False
            if "import sqlalchemy_utils" not in data:
                data = "import sqlalchemy_utils\n" + data
                modified = True
            if "import sqlalchemy_utils.types" not in data:
                data = "import sqlalchemy_utils.types\n" + data
                modified = True
            if modified:
                fn.write_text(data)

    def should_run(self):
        return True


class UpdateIndexWizardStep(ModelWizardStep):
    steps = (
        RadioWizardStep(
            "update_opensearch",
            options={
                "run": f"{Fore.GREEN}Update opensearch index{Style.RESET_ALL}",
                "skip": f"{Fore.RED}Do not update opensearch index{Style.RESET_ALL}",
            },
            default="run",
            heading=f"""
Before the model can be used, I need to create index inside opensearch server.
This is not necessary if the model has not been changed. Should I create/update
the index? 
                            """,
            force_run=True,
        ),
    )

    def after_run(self):
        setup_cfg = read_configuration(self.model_dir / "setup.cfg")
        for cmd_def in setup_cfg["options"]["entry_points"]["flask.commands"]:
            cmd_name, _ = [x.strip() for x in cmd_def.split("=", maxsplit=1)]
            break

        if self.data["update_opensearch"] == "run":
            self.invenio_command(cmd_name, "reindex", "--recreate")

    def should_run(self):
        return True



from __future__ import annotations

from oarepo_cli.ui.wizard import WizardStep
from oarepo_cli.ui.wizard.steps import InputWizardStep
from oarepo_cli.utils import commit_git


class PrimarySiteNameStep(WizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            steps=[
                InputWizardStep(
                    "primary_site_name",
                    prompt="""Directory name of your site (keep the default if unsure)""",
                    default=lambda data: data["project_package"].replace("_", "-")
                    + "-site",
                )
            ],
            **kwargs,
        )

    def after_run(self):
        super().after_run()
        commit_git(
            self.data.project_dir,
            "after-primary-site-name",
            "Committed automatically after primary site name has been selected",
        )

    def should_run(self):
        return "primary_site_name" not in self.data



from __future__ import annotations

from oarepo_cli.cli.site.utils import SiteWizardStepMixin
from oarepo_cli.ui.wizard import WizardStep


class FixupSiteCodeStep(SiteWizardStepMixin, WizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            heading="""
Current invenio sources have some hard-coded values in subsequent checks.
In this step I'll patch them up to use the current invenio config.
""",
            **kwargs,
        )

    def get_service_health_file(self):
        invenio_cli_venv = (
            self.data.project_dir / self.data.get("config.invenio_cli")
        ).parent.parent
        for candidate in invenio_cli_venv.rglob("services_health.py"):
            if (
                candidate.parent.name == "commands"
                and candidate.parent.parent.name == "invenio_cli"
            ):
                return candidate
        raise Exception(
            "Can not find services_health.py - it seems that Invenio has changed its source files structure. Please report this issue to du-suppport@cestnet.cz"
        )

    def after_run(self):
        health_file = self.get_service_health_file()
        contents = health_file.read_text()
        contents = contents.replace(
            '"localhost:9200/_cluster/health?wait_for_status=yellow"',
            "oarepo_elasticsearch_patch()",
        )
        contents = (
            contents
            + """
def oarepo_elasticsearch_patch():
    from pathlib import Path
    import re
    cwd = Path.cwd()
    invenio_cfg = cwd / 'invenio.cfg'
    if not invenio_cfg.exists():
        raise Exception(f'Please run the check script from within the site directory. Current directory is {cwd}')
    loading = False
    lines = []
    for l in invenio_cfg.read_text().split('\\n'):
        if loading:
            lines.append(l)
            if not l.startswith(' '):
                break
        if l.startswith('SEARCH_HOSTS'):
            loading = True
            lines.append(l)
    g = {}
    lines[0] = re.sub('SEARCH_HOSTS\\s*=', '', lines[0])
    # eval is nasty, but here it should be ok - the invenio.cfg was generated by us
    ret = eval('\\n'.join(lines), g, g)
    search_decl = ret[0]
    return f"{search_decl['host']}:{search_decl['port']}/_cluster/health?wait_for_status=yellow"        """
        )
        health_file.write_text(contents)

    def should_run(self):
        health_file = self.get_service_health_file()
        contents = health_file.read_text()
        return "oarepo_elasticsearch_patch" not in contents



from __future__ import annotations

import shutil
import venv

from oarepo_cli.ui.wizard import WizardStep

from ...utils import commit_git, pip_install


class InstallIOARepoCliStep(WizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            heading="""
I will install oarepo command-line tools that make using the invenio easier.
To run them, invoke the "nrp-cli" script from within the project directory.            
            """,
            **kwargs,
        )

    def after_run(self):
        print("Creating nrp-cli virtualenv")
        oarepo_cli_dir = self._oarepo_cli_dir
        self.data["oarepo_cli"] = str(
            (oarepo_cli_dir / "bin" / "nrp-cli").relative_to(self.data.project_dir)
        )
        if oarepo_cli_dir.exists():
            shutil.rmtree(oarepo_cli_dir)
        venv.main([str(oarepo_cli_dir)])

        pip_install(
            oarepo_cli_dir / "bin" / "pip",
            "OAREPO_CLI_VERSION",
            "oarepo-cli>=11.0.8,<12",
            "https://github.com/oarepo/oarepo-cli",
        )

        with open(oarepo_cli_dir / ".check.ok", "w") as f:
            f.write("oarepo check ok")
        commit_git(
            self.data.project_dir,
            "after-install-oarepo-cli",
            "Committed automatically after oarepo-cli has been installed",
        )

    @property
    def _oarepo_cli_dir(self):
        return self.data.project_dir / ".venv" / "nrp-cli"

    def should_run(self):
        return not (self._oarepo_cli_dir / ".check.ok").exists()



import click as click

from oarepo_cli.cli.model.utils import ProjectWizardMixin
from oarepo_cli.cli.utils import PipenvInstallWizardStep, SiteMixin, with_config
from oarepo_cli.ui.wizard import Wizard
from oarepo_cli.ui.wizard.steps import WizardStep


@click.command(
    name="install",
    help="""
    Install the UI to the site. Required arguments:
    <name>   ... name of the ui. The recommended pattern for it is <modelname>-ui
    """,
)
@click.argument("name")
@with_config(config_section=lambda name, **kwargs: ["ui", name])
def install_ui(cfg=None, **kwargs):
    InstallWizard().run(cfg)


class InstallWizardStep(PipenvInstallWizardStep):
    folder = "ui"


class CompileAssetsStep(SiteMixin, ProjectWizardMixin, WizardStep):
    def should_run(self):
        return True

    def after_run(self):
        self.invenio_cli_command("assets", "build")


class InstallWizard(ProjectWizardMixin, Wizard):
    steps = [InstallWizardStep(), CompileAssetsStep()]

    def should_run(self):
        return super().should_run()



from __future__ import annotations


def required(prop_name, prop_label=None):
    def _val(data):
        v = data.get(prop_name)
        if v is None or v == "":
            return f"Value required"

    return _val



import click as click

from oarepo_cli.cli.model.utils import ModelWizardStep
from oarepo_cli.cli.utils import with_config
from oarepo_cli.ui.wizard import Wizard


@click.command(
    name="load",
    help="""
Import (sample) data. Required arguments:
    <name>       ... name of the already existing model
    <data-path>  ... path to the data. If not filled, sample data will be used
    """,
)
@click.argument("name", required=False)
@click.argument(
    "data-path", required=False, type=click.Path(file_okay=True, dir_okay=False)
)
@with_config(config_section=lambda name, **kwargs: ["models", name])
def load_data(cfg=None, data_path=None, *args, **kwargs):
    if not data_path:
        data_path = cfg.project_dir / cfg["model_dir"] / "data" / "sample_data.yaml"
        cfg["data_path"] = "data" / "sample_data.yaml"
    else:
        cfg["data_path"] = data_path

    w = Wizard(ImportDataWizardStep())
    w.run(cfg)


class ImportDataWizardStep(ModelWizardStep):
    def after_run(self):
        data_path = self.model_dir.join(self.data["data_path"])
        self.invenio_command(self.model_name, "load", data_path)



import sys
import time


def slow_print(s):
    wc = 0
    for c in s:
        print(c, end="")
        sys.stdout.flush()
        wc += 1
        time.sleep(0.003)
        if c == "\n" and wc > 2:
            time.sleep(0.003 * max(1, 80 - wc))
            wc = 0
    print()



import click as click

from .add import add_ui

# from .compile import compile_ui
from .install import install_ui
from .list import list_uis


@click.group(help="User interface related tools (add user interface for a model, ...)")
def ui():
    pass


ui.add_command(add_ui)
# ui.add_command(compile_ui)
ui.add_command(install_ui)
ui.add_command(list_uis)



import json
import click

from oarepo_cli.cli.utils import with_config
from oarepo_cli.utils import run_cmdline


@click.command
@with_config()
@click.pass_context
def upgrade(ctx, project_dir, cfg, **kwargs):
    "Upgrade all virtualenvs in this repository"
    for venv_dir in (project_dir / ".venv").glob("*"):
        if not venv_dir.is_dir():
            continue
        if not (venv_dir / "bin" / "python").exists():
            continue
        upgrade_venv(venv_dir)
    for site in cfg.whole_data["sites"].values():
        upgrade_site(project_dir / site["site_dir"])


def upgrade_venv(venv_dir):
    # run
    packages = run_cmdline(
        "./bin/pip",
        "list",
        "--outdated",
        "--format",
        "json",
        cwd=venv_dir,
        grab_stdout=True,
        grab_stderr=False,
        raise_exception=True,
    )
    packages = json.loads(packages)
    obsolete_packages = [
        f"{p['name']}=={p['latest_version']}"
        for p in packages
        if p["name"].startswith("oarepo") or p["name"].startswith("nrp")
    ]
    if obsolete_packages:
        run_cmdline(
            "./bin/pip",
            "install",
            "-U",
            *obsolete_packages,
            cwd=venv_dir,
            raise_exception=True,
        )


def upgrade_site(site_dir):
    venv_dir = site_dir / ".venv"
    if not venv_dir.exists():
        venv_dir.mkdir(parents=True)
    run_cmdline(
        "pipenv",
        "update",
        cwd=site_dir,
        environ={"PIPENV_IGNORE_VIRTUALENVS": "1"},
    )



import click as click
from oarepo_cli.cli.utils import with_config
import subprocess
from oarepo_cli.config import MonorepoConfig
import time
import os


@click.command(
    name="develop",
    hidden=True,
    help="Start a development docker-compose",
)
@click.option("--site", help="Name of the site to start")
@click.option("--rebuild/--no-rebuild", help="Rebuild the container", default=True)
@click.option(
    "--nrp-cli-directory",
    help="Use this directory for nrp-cli (nrp-cli development only)",
    hidden=True,
)
@with_config()
def develop(
    cfg: MonorepoConfig = None,
    site=None,
    rebuild=True,
    nrp_cli_directory=None,
    **kwargs,
):
    if not site:
        site = list(cfg.whole_data["sites"].keys())[0]

    if rebuild:
        with open(cfg.project_dir / "sites" / site / "requirements.txt", "w") as file:
            subprocess.check_call(
                [
                    "pipenv",
                    "requirements",
                ],
                cwd=cfg.project_dir / "sites" / site,
                stdout=file,
                env={**os.environ, "PIPENV_IGNORE_VIRTUALENVS": "1"},
            )
        subprocess.check_call(
            [
                "docker",
                "build",
                ".",
                "-f",
                f"sites/{site}/development/Dockerfile",
                "-t",
                f"{site}:devel",
                "--build-arg",
                f"REPOSITORY_SITE_NAME={site}",
                "--no-cache",
                "--progress",
                "plain",
            ],
            cwd=cfg.project_dir,
        )

    # run the services
    subprocess.Popen(
        [
            "docker",
            "compose",
            "-f",
            "docker-compose.development.yml",
            "up",
            "-d",
            "cache",
            "db",
            "mq",
            "search",
            "s3",
        ],
        cwd=cfg.project_dir / "sites" / site,
    )
    # wait a bit, proper checking for started containers should be from the app service
    time.sleep(5)
    print(
        "Please make sure that the containers (cache, db, mq, search, s3) are up and running. "
        "The future version will check for this automatically. Press Enter when ok"
    )
    input()

    app = subprocess.call(
        [
            "docker",
            "compose",
            "-f",
            "docker-compose.development.yml",
            "run",
            "-i",
            "app",
        ],
        cwd=cfg.project_dir / "sites" / site,
    )



import subprocess
import venv
from pathlib import Path

import click as click

from oarepo_cli.cli.model.utils import ModelWizardStep
from oarepo_cli.cli.utils import with_config
from oarepo_cli.ui.wizard import InputWizardStep, StaticWizardStep, Wizard, WizardStep
from oarepo_cli.ui.wizard.steps import RadioWizardStep
from oarepo_cli.utils import get_cookiecutter_source, pip_install, to_python_name

from ...utils import commit_git


@click.command(
    name="add",
    help="""
Generate a new model. Required arguments:
    <name>   ... name of the model, can contain [a-z] and dash (-)""",
)
@click.argument("name", required=True)
@click.option(
    "--merge",
    multiple=True,
    help="""
Use this option to merge your code into the generated model.

--merge my_dir              will merge my_dir with the generated sources

--merge my_dir=gen_subdir   will merge my_dir into the subdir(relative path to models/<model>)

--merge my_file=<rel_path_to_file>   will merge single file

Normally, user file is merged at the end of the generated file - that is, the content of generated file goes first (includes, classes, arrays).

Use '-' before the dir/file to reverse the order - the content of your file will be prepended to an existing file
Use '!' before the dir/file to copy the file to destination without merging it
""",
)
@with_config(config_section=lambda name, **kwargs: ["models", name])
def add_model(cfg=None, merge=None, **kwargs):
    commit_git(
        cfg.project_dir,
        f"before-model-add-{cfg.section}",
        f"Committed automatically before model {cfg.section} has been added",
    )
    if merge:
        venv_dir: Path = cfg.project_dir / ".venv" / "oarepo-model-builder"
        venv_dir = venv_dir.absolute()
        if not venv_dir.exists():
            venv_dir.parent.mkdir(parents=True, exist_ok=True)
            venv.main([str(venv_dir)])

            pip_install(
                venv_dir / "bin" / "pip",
                "OAREPO_MODEL_BUILDER_VERSION",
                "oarepo-model-builder==3.*",
                "https://github.com/oarepo/oarepo-model-builder",
            )

    add_model_wizard.run(cfg)

    if merge:
        for merge_def in merge:
            opts = []
            merge_def = merge_def.split("=", maxsplit=1)

            merge_source: Path = merge_def[0]
            if merge_source[0] == "-":
                merge_source = merge_source[1:]
                opts.append("--destination-first")
            if merge_source[0] == "!":
                merge_source = merge_source[1:]
                opts.append("--overwrite")

            merge_target: Path = cfg.project_dir
            for p in cfg.section_path:
                merge_target = merge_target / p
            if len(merge_def) == 2:
                merge_target = merge_target.joinpath(merge_def[1])

            subprocess.call(
                [
                    venv_dir / "bin" / "oarepo-merge",
                    Path(merge_source).absolute(),
                    Path(merge_target).absolute(),
                    *opts,
                ]
            )
    commit_git(
        cfg.project_dir,
        f"after-model-add-{cfg.section}",
        f"Committed automatically after model {cfg.section} has been added",
    )


class CreateModelWizardStep(ModelWizardStep, WizardStep):
    def after_run(self):
        model_dir = self.model_dir
        base_model_package = {
            "empty": "(none)",
            "common": "nr-common-metadata-model-builder",
            "documents": "nr-documents-records-model-builder",
            "data": "TODO",
        }.get(self.data["model_kind"])
        base_model_use = base_model_package.replace("-model-builder", "")

        cookiecutter_path, cookiecutter_branch = get_cookiecutter_source(
            "OAREPO_MODEL_COOKIECUTTER_VERSION",
            "https://github.com/oarepo/cookiecutter-model",
            "v11.0",
            master_version="master",
        )

        self.run_cookiecutter(
            template=cookiecutter_path,
            config_file=f"model-{model_dir.name}",
            checkout=cookiecutter_branch,
            output_dir=str(model_dir.parent),
            extra_context={
                **self.data,
                "model_name": model_dir.name,
                "base_model_package": base_model_package,
                "base_model_use": base_model_use,
            },
        )
        self.data["model_dir"] = str(model_dir.relative_to(self.data.project_dir))

    def should_run(self):
        return not self.model_dir.exists()


add_model_wizard = Wizard(
    StaticWizardStep(
        heading="""
Before creating the datamodel, I'll ask you a few questions.
If unsure, use the default value.
    """,
    ),
    InputWizardStep(
        "model_package",
        prompt="Enter the model package",
        default=lambda data: to_python_name(data.section),
    ),
    RadioWizardStep(
        "model_kind",
        heading="""
Now choose if you want to start from a completely empty model or if you
want to base your model on an already existing one. You have the following
options:

* common       - a common set of metadata created by the National library of Technology, Prague
                 compatible with Dublin Core
                 See https://github.com/Narodni-repozitar/nr-common-metadata for details
* documents    - extension of common, can be used to capture metadata of documents (articles etc.)
                 See https://github.com/Narodni-repozitar/nr-documents-records for details
* data         - extension of common for capturing generic metadata about datasets
                 See TODO for details
* custom_model - use any custom model as a base model. If you select this option, answer the next two questions
                 (base_model_package, base_model_use) as well
* empty_model  - just what it says, not recommended as you have no compatibility with
                 the Czech National Repository

When asked about the base_model_package: leave as is unless you have chosen a custom base model.
In that case enter the package name of the model builder extension on pypi which contains the custom model.

When asked about the base_model_use: leave as is unless you have chosen a custom base model.
In that case enter the string that should be put to 'oarepo:use. Normally that is the name
of the extension without 'model-builder-'. See the documentation of your custom model for details.
    """,
        options={
            "common": "Common set of metadata, DC compatible",
            "documents": "Based on Czech National Repository documents metadata",
            "data": "Based on Czech National Repository datasets metadata",
            "empty": "Just use empty model, I'll add the metadata myself",
        },
        default="common",
    ),
    StaticWizardStep(
        heading="""
Now tell me something about you. The defaults are taken from the monorepo, feel free to use them.
    """,
    ),
    InputWizardStep(
        "author_name",
        prompt="""Model author""",
        default=lambda data: (get_site(data) or {}).get("author_name"),
    ),
    InputWizardStep(
        "author_email",
        prompt="""Model author's email""",
        default=lambda data: (get_site(data) or {}).get("author_email"),
    ),
    RadioWizardStep(
        "permissions_presets",
        heading="Would you like to set up permissions",
        options={
            "no": "no",
            "read_only": "read only access to records, no one can create or edit them",
            "everyone": "every user (both authenticated or anonymous) can create, edit and delete any record",
        },
        default="yes",
    ),
    InputWizardStep(
        "pid_type",
        prompt="""Specify the name of the type of the pid of the model. If nothing is provided,
         it will be autogenerated.""",
        default="",
        required=False,
    ),
    RadioWizardStep(
        "use_drafts",
        heading="Should records be first uploaded as editable drafts before they are published?",
        options={
            "yes": "yes",
            "no": "no",
        },
        default="yes",
    ),
    StaticWizardStep(
        heading="Now you can choose which plugins you need in the repo.", pause=True
    ),
    RadioWizardStep(
        "use_files",
        heading="Will you upload files to the records in this model? If the repository is not metadata only, answer yes.",
        options={
            "yes": "yes",
            "no": "no",
        },
        default="yes",
    ),
    RadioWizardStep(
        "use_requests",
        heading="Do you need approval process for the records in this model? We recommend to use it, otherwise your changes would be immediately visible.",
        options={
            "yes": "yes",
            "no": "no",
        },
        default="yes",
    ),
    RadioWizardStep(
        "use_expandable_fields",
        heading="Will you use expandable fields? If in doubt, choose no.",
        options={
            "yes": "yes",
            "no": "no",
        },
        default="no",
    ),
    RadioWizardStep(
        "use_relations",
        heading="Will you use relations to different records? For example, if this model represents a Dataset, relation to Article model.",
        options={
            "yes": "yes",
            "no": "no",
        },
        default="yes",
    ),
    RadioWizardStep(
        "use_custom_fields",
        heading="Would you like to make your model extensible during deployment via custom fields?",
        options={
            "yes": "yes",
            "no": "no",
        },
        default="yes",
    ),
    RadioWizardStep(
        "use_vocabularies",
        heading="Will you use vocabularies in your model?",
        options={
            "yes": "yes",
            "no": "no",
        },
        default="yes",
    ),
    StaticWizardStep(
        heading="Now I have all the information to generate your model. After pressing Enter, I will generate the sources",
        pause=True,
    ),
    CreateModelWizardStep(),
    StaticWizardStep(
        heading=lambda data: f"""
The model has been generated in the {data.section} directory.
At first, edit the metadata.yaml and then run "nrp-cli model compile {data.section}"
and to install to the site run "nrp-cli model install {data.section}".
                     """,
        pause=True,
    ),
)


def get_site(data):
    primary_site_name = data.get("config.primary_site_name")
    return data.get(f"sites.{primary_site_name}")



import subprocess

from oarepo_cli.cli.site.utils import SiteWizardStepMixin
from oarepo_cli.ui.wizard import WizardStep

from ...utils import run_cmdline


class InitFilesStep(SiteWizardStepMixin, WizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            heading="""
        Now I will configure the default location for files storage in the minio s3 framework.
            """,
            **kwargs,
        )

    def after_run(self):
        from minio import Minio

        client = Minio(
            "localhost:9000",
            access_key="CHANGE_ME",
            secret_key="CHANGE_ME",
            secure=False,
        )
        bucket_name = self.data["site_package"].replace(
            "_", ""
        )  # bucket names with underscores are not allowed
        if not client.bucket_exists(bucket_name):
            client.make_bucket(bucket_name)
        run_cmdline(
            "pipenv",
            "run",
            "invenio",
            "files",
            "location",
            "default",
            f"s3://{bucket_name}",
            "--default",
            cwd=self.site_dir,
            environ={"PIPENV_IGNORE_VIRTUALENVS": "1"},
        )
        self.check_file_location_initialized(raise_error=True)

    def check_file_location_initialized(self, raise_error=False):
        try:
            output = run_cmdline(
                "pipenv",
                "run",
                "invenio",
                "files",
                "location",
                "list",
                cwd=self.site_dir,
                environ={"PIPENV_IGNORE_VIRTUALENVS": "1"},
                grab_stdout=True,
                raise_exception=True,
            )
            print(f"initialization check:\n{output}\n")
        except subprocess.CalledProcessError:
            raise Exception("Checking if file location exists failed.")
        if output:
            return True
        else:
            if raise_error:
                raise Exception(
                    "No file location exists. This probably means that the wizard was unable to create one."
                )
            return False

    def should_run(self):
        return not self.check_file_location_initialized(raise_error=False)



from io import StringIO
from pathlib import Path

import yaml
from deepmerge import always_merger


class Config:
    def __init__(self):
        self.config = {}

    def __getitem__(self, item):
        return self.config[item]

    def __setitem__(self, key, value):
        self.config[key] = value
        self.on_changed()

    def __iter__(self):
        return iter(self.config)

    def keys(self):
        return self.config.keys()

    def values(self):
        return self.config.values()

    def items(self):
        return self.config.items()

    def get(self, item, default=None):
        return self.config.get(item, default)

    def setdefault(self, item, default):
        return self.config.setdefault(item, default)

    def on_changed(self):
        pass


class MonorepoConfig(Config):
    type = "monorepo"

    def __init__(self, path: Path, section=["config"]):
        super().__init__()
        self.path = path
        self.existing = False
        if not section:
            section = []
        elif isinstance(section, str):
            section = [section]
        self.section_path = tuple(section)
        self.whole_data = {}
        self._load_section()

    def load(self):
        with open(self.path, "r") as f:
            data = yaml.safe_load(f)
            self.whole_data = data
            self._load_section()
            self.existing = True

    def _load_section(self):
        data = self.whole_data
        for s in self.section_path:
            data = data.setdefault(s, {})
        self.config = data

    def save(self):
        data = {**self.whole_data, "type": self.type}
        # just try to dump so that if that is not successful we do not overwrite the config
        sio = StringIO()
        yaml.safe_dump(data, sio)

        # and real dump here
        if self.path.parent.exists():
            with open(self.path, "w") as f:
                f.write(sio.getvalue())

    def on_changed(self):
        if self.path.parent.exists():
            self.save()

    def _section(self, name, default=None):
        name = name.split(".")
        d = self.whole_data
        for n in name[:-1]:
            d = d.get(n, {})
        return d.get(name[-1], default)

    def get(self, item, default=None):
        if "." in item:
            return self._section(item, default)
        return super().get(item, default)

    @property
    def section(self):
        return self.section_path[-1]

    @property
    def project_dir(self):
        return self.path.parent.resolve()

    def merge_config(self, config_data, top=False):
        if top:
            always_merger.merge(self.whole_data, config_data)
            self._load_section()
            self.save()
        else:
            always_merger.merge(self.config, config_data)



from oarepo_cli.cli.site.utils import SiteWizardStepMixin
from oarepo_cli.ui.wizard import WizardStep

from ...utils import commit_git, run_cmdline


class InstallInvenioStep(SiteWizardStepMixin, WizardStep):
    def __init__(self, **kwargs):
        super().__init__(
            heading="""
Now I'll install invenio site.

Note that this can take a lot of time as UI dependencies
will be downloaded and installed and UI will be compiled.
            """,
            **kwargs,
        )

    def after_run(self):
        run_cmdline(
            self.data.project_dir / self.data.get("config.invenio_cli"),
            "install",
            cwd=self.site_dir,
            environ={"PIPENV_IGNORE_VIRTUALENVS": "1"},
        )
        if not self._manifest_file.exists():
            raise FileNotFoundError(
                "invenio-cli install has not created var/instance/static/dist/manifest.json."
                "Please check the output, correct errors and run this command again"
            )
        commit_git(
            self.data.project_dir,
            f"after-site-invenio-cli-install-{self.data.section}",
            f"Committed automatically after site {self.data.section} has been installed",
        )

    @property
    def _pipenv_venv_dir(self):
        success = run_cmdline(
            "pipenv",
            "--venv",
            cwd=self.site_dir,
            environ={"PIPENV_IGNORE_VIRTUALENVS": "1"},
            check_only=True,
            grab_stdout=True,
        )
        if not success:
            return None
        return success.strip()

    def should_run(self):
        manifest_file = self._manifest_file
        return not manifest_file.exists()

    @property
    def _manifest_file(self):
        pipenv_dir = self.data.project_dir / self.data["site_pipenv_dir"]
        manifest_file = (
            pipenv_dir / "var" / "instance" / "static" / "dist" / "manifest.json"
        )

        return manifest_file
